@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt.h - branch tracer for linux common header                             */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -31,12 +31,17 @@
 #  include <linux/version.h>
 #  include <linux/file.h>
 #  include <linux/proc_fs.h>
+#  if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+#    include <linux/irq.h>
+#  endif
 #else
 #  include <sys/types.h>
+#  ifdef HAVE_CONFIG_H
+#    include "config.h"
+#  endif
 #endif
 
 #define MOD_NAME		"bt_mod"
-#define BT_VER			"1.0.2"
 
 #define BT_FLAG_T_START		(1 << 26)
 #define BT_FLAG_T_STOP		(1 << 27)
@@ -44,6 +49,7 @@
 #define BT_FLAG_DEBUG		(1 << 29)
 #define BT_FLAG_COMM		(1 << 30)
 #define BT_FLAG_PID		(1 << 31)
+#define	BT_FLAG_PID_LL		((unsigned long long)BT_FLAG_PID << 32)
 #define	is_bt_record(r)		\
 	(!((r)->flags & (BT_FLAG_WARN|BT_FLAG_DEBUG|BT_FLAG_PID)))
 #define	is_pid_record(r)	((r)->flags & BT_FLAG_PID)
@@ -56,29 +62,28 @@ struct bt_record {
 	unsigned long	from;
 	unsigned long	to;
 	unsigned long	flags;
-};
+} __attribute__((packed));
 
 struct pid_record {
-	unsigned long	pid;
-	unsigned long	tv_sec;
-	unsigned long	tv_usec;
-};
+	unsigned long		pid;
+	unsigned long long	clocks;
+} __attribute__((packed));
 
 struct comm_record {
 	char		comm[BT_COMM_LEN];
 	unsigned long	flags;
-};
+} __attribute__((packed));
 
 struct warn_record {
 	unsigned long	left;
 	unsigned long	reserve;
 	unsigned long	flags;
-};
+} __attribute__((packed));
 
 struct tmr_record {
 	unsigned long long	timestamp;
 	unsigned long		flags;
-};
+} __attribute__((packed));
 
 struct ds_manage {
 	unsigned long	bts_base;
@@ -118,7 +123,9 @@ struct info_per_cpu {
 #include <asm/types.h>
 
 /* MSRs */
-#define MSR_IA32_DS_AREA	0x600
+#ifndef MSR_IA32_DS_AREA
+#  define MSR_IA32_DS_AREA	0x600
+#endif
 #define MSR_DEBUGCTL		0x1d9
 struct debugctl_bits {
 	int	lbr;
@@ -190,6 +197,14 @@ typedef struct hw_interrupt_type bt_int_t;
 #define	bt_d_child d_child
 #endif
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+typedef unsigned long (*kallsyms_lookup_name_t)(const char *name);
+typedef void (*set_irq_chip_and_handler_t)(unsigned int irq,
+					   struct irq_chip *chip,
+					   irq_flow_handler_t handle);
+typedef void (*handle_irq_t)(unsigned int irq, struct irq_desc *desc);
+#endif
+
 static inline void chk_procs_using_relfs(struct dentry *d,
 					 pid_t **pp_pid, pid_t *p_max)
 {
