@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_hook.c - branch trace module (hook interface)                         */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -50,7 +50,6 @@ extern unsigned long syscall_filter_tbl[];
 
 pid_t bt_collect_pid;
 
-void write_pid_record(pid_t);
 void bts_log_write(void);
 int is_trace_pid(pid_t, unsigned long*);
 void add_pid_tbl(pid_t, unsigned long*);
@@ -73,12 +72,21 @@ void check_and_wait_relfs_write(void);
  * child process's pid_tbl check could be faster than parent process.
  * For this case, we also check the parent pid.
  */
-#define is_target_user_process(p, pid_tbl)	\
+#if (RHEL_MVER == 5)
+  #define is_target_user_process(p, pid_tbl)	\
+	(!is_kernel_thread(p) \
+	 && ((is_syscall(mode) \
+	      && (is_trace_pid((p)->parent->pid, pid_tbl) \
+		  || is_trace_pid((p)->group_leader->pid, pid_tbl))) \
+	     || is_trace_pid((p)->pid, pid_tbl)))
+#else
+  #define is_target_user_process(p, pid_tbl)	\
 	(!is_kernel_thread(p) \
 	 && ((is_syscall(mode) \
 	      && (is_trace_pid((p)->real_parent->pid, pid_tbl) \
 		  || is_trace_pid((p)->group_leader->pid, pid_tbl))) \
 	     || is_trace_pid((p)->pid, pid_tbl)))
+#endif
 
 static void context_switch_probe(struct pt_regs *regs)
 {
@@ -124,7 +132,7 @@ EXIT:
 	put_cpu();
 }
 
-static void sysenter_probe(const ctr_handler_t h)
+static void sysenter_probe(const ctr_arg_t arg)
 {
 	int cpu, save;
 	long n_syscall;
@@ -134,7 +142,7 @@ static void sysenter_probe(const ctr_handler_t h)
 	cpu = get_cpu();
 	if (!cpu_isset(cpu, enable))
 		goto EXIT;
-	n_syscall = h.regs.orig_eax;
+	n_syscall = arg.regs.orig_eax;
 	/*
 	serial_prints("ENT:%d:%d:%d\n",
 		      smp_processor_id(), current->pid, n_syscall);
@@ -173,7 +181,7 @@ EXIT:
 	put_cpu();
 }
 
-static void sysexit_probe(const ctr_handler_t h)
+static void sysexit_probe(const ctr_arg_t arg)
 {
 	int cpu, save;
 	long n_syscall;
@@ -183,7 +191,7 @@ static void sysexit_probe(const ctr_handler_t h)
 	cpu = get_cpu();
 	if (!cpu_isset(cpu, enable))
 		goto EXIT;
-	n_syscall = h.regs.orig_eax;
+	n_syscall = arg.regs.orig_eax;
 	/*
 	serial_prints("EXT:%d:%d:%d\n",
 		      smp_processor_id(), current->pid, n_syscall);
@@ -197,9 +205,9 @@ static void sysexit_probe(const ctr_handler_t h)
 		if (n_syscall == 120) { /* sys_clone? */
 			/*
 			serial_prints("ADD:%d->%d\n",
-				      current->pid, h.regs.eax);
+				      current->pid, arg.regs.eax);
 				      */
-			add_pid_tbl(h.regs.eax, pid_tbl);
+			add_pid_tbl(arg.regs.eax, pid_tbl);
 		}
 		if (!is_trace_syscall(n_syscall)) {
 			bts_facility_restore(save);
