@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_irq.c - branch trace module (interrupt handler)                       */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -33,6 +33,10 @@ extern struct debugctl_bits ctl_bits;
 extern DEFINE_PER_CPU(struct info_per_cpu, bt_info_per_cpu);
 extern unsigned long irq_desc_p;
 extern unsigned long no_irq_type_p;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+extern set_irq_chip_and_handler_t set_irq_chip_and_handler_f;
+extern handle_irq_t handle_edge_irq_f;
+#endif
 
 static char *irq_name = "bts";
 static int pmi_irq = -1;
@@ -118,11 +122,22 @@ static struct hw_interrupt_type	bt_int_type = {
 	.set_affinity = set_affinity_pmi
 };
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+irqreturn_t bt_interrupt(int irq, void *dev_id)
+#else
 irqreturn_t bt_interrupt(int irq, void *dev_id, struct pt_regs *regs)
+#endif
 {
 	return IRQ_HANDLED;
 }
 
+/* Here are the macros for the old kernel. */
+#ifndef IRQF_DISABLED
+#  define IRQF_DISABLED		SA_INTERRUPT
+#endif
+#ifndef IRQF_SHARED
+#  define IRQF_SHARED		SA_SHIRQ
+#endif
 static int init_isr_per_system(void)
 {
 	int irq;
@@ -137,8 +152,12 @@ static int init_isr_per_system(void)
 	if (irq == 15)
 		return -EBUSY;
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+	set_irq_chip_and_handler_f(irq, &bt_int_type, handle_edge_irq_f);
+#else
 	desc[irq].bt_handler = &bt_int_type;
-	rc = request_irq(irq, bt_interrupt, SA_INTERRUPT | SA_SHIRQ,
+#endif
+	rc = request_irq(irq, bt_interrupt, IRQF_DISABLED | IRQF_SHARED,
 			 irq_name, irq_name);
 	if (rc) {
 		printk("%s: cannot setup irq %d for BTS\n", MOD_NAME, irq);
