@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_main.h - branch trace module header                                   */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -42,16 +42,25 @@ unsigned int ctl_on_bits, ctl_off_bits;
 cpumask_t enable;
 unsigned long *pid_tbl;
 unsigned long *syscall_pid_tbl;
-int pid_max;
+int pid_max = PID_MAX_DEFAULT;
 ////spinlock_t pid_tbl_lock = SPIN_LOCK_UNLOCKED;
 DECLARE_BITMAP(syscall_exe_tbl, NR_syscalls);
 DECLARE_BITMAP(syscall_filter_tbl, NR_syscalls);
 
+unsigned long idt_table_p;
 unsigned long irq_desc_p;
 unsigned long no_irq_type_p;
 unsigned long switch_to_addr;
 int switch_to_size;
 
+unsigned long kallsyms_lookup_name_addr;
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+kallsyms_lookup_name_t kallsyms_lookup_name_f;
+/* below variables are got by kallsyms_lookup_name */
+set_irq_chip_and_handler_t set_irq_chip_and_handler_f;
+handle_irq_t handle_edge_irq_f;
+#endif
+
 int chk_sctime;
 
 size_t btsbuf_size = BTS_BUF_MIN_SIZE;
@@ -60,6 +69,7 @@ size_t subbuf_size = 2 * 1024 * 1024;
 size_t subbuf_num = 16;
 size_t subbuf_sleep_threshold = 8;
 int mode;
+unsigned long long start_clocks;
 
 DEFINE_PER_CPU(struct info_per_cpu, bt_info_per_cpu);
 
@@ -390,7 +400,7 @@ static int check_bts_availability(void)
 {
 	unsigned int eax, ebx, ecx, edx;
 	unsigned char family, model;
-	
+
 	cpuid(0x01, &eax, &ebx, &ecx, &edx);
 	if (!(edx & (1 << 21))) {
 		printk("%s: BTS facility is not available in this processor.\n",
@@ -403,12 +413,12 @@ static int check_bts_availability(void)
 		family += (eax >> 20) & 0xff;
 		model += ((eax >> 16) & 0xf) << 4;
 	}
-	if (family == 6 && (model == 9 || model == 13)) /* cpu is Pentium-M? */
+	/* cpu is Pentium-M or Core Solo/Duo? */
+	if (family == 6 && (model == 9 || model == 13 || model == 14))
 		ctl_bits = ctl_bits_pm;
 	else
 		ctl_bits = ctl_bits_p4;
-	ctl_on_bits  = ctl_bits.bts|ctl_bits.tr|ctl_bits.btint;
-	ctl_off_bits = ctl_bits.bts|ctl_bits.tr|ctl_bits.btint|ctl_bits.lbr;
+	ctl_on_bits = ctl_off_bits = ctl_bits.bts|ctl_bits.tr|ctl_bits.btint;
 
 	rdmsr(MSR_DEBUGCTL, eax, edx);
 	if (eax & ctl_bits.bts) {
@@ -430,7 +440,7 @@ static void setup_ds_area_per_cpu(void *data)
 	if (!ds_mng)
 		return;
 	rdmsr(MSR_IA32_DS_AREA, low, high);
-	wrmsr(MSR_IA32_DS_AREA, ds_mng, high);
+	wrmsr(MSR_IA32_DS_AREA, (unsigned int)ds_mng, high);
 }
 
 static int setup_ds_area(void)
@@ -481,13 +491,30 @@ static int bt_mod_init(void)
 	int rc = 0;
 
 	/* check module parameters */
-	if (!irq_desc_p || !no_irq_type_p || !switch_to_addr || !switch_to_size
+	if (!idt_table_p || !irq_desc_p || !no_irq_type_p
+	    || !switch_to_addr || !switch_to_size
 	    || !subbuf_num || subbuf_size < BTS_BUF_MIN_SIZE
 	    || !subbuf_sleep_threshold || btsbuf_size < BTS_BUF_MIN_SIZE
 	    || bts_int_margin_recs < MIN_INT_MARGIN_RECS) {
 		rc = -EINVAL;
 		goto ERROR;
 	}
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
+	if (!kallsyms_lookup_name_addr) {
+		rc = -EINVAL;
+		goto ERROR;
+	}
+	kallsyms_lookup_name_f =
+		(kallsyms_lookup_name_t)kallsyms_lookup_name_addr;
+	set_irq_chip_and_handler_f = (set_irq_chip_and_handler_t)
+		kallsyms_lookup_name_f("set_irq_chip_and_handler");
+	handle_edge_irq_f = (handle_irq_t)
+		kallsyms_lookup_name_f("handle_edge_irq");
+	if (!set_irq_chip_and_handler_f || !handle_edge_irq_f) {
+		rc = -EINVAL;
+		goto ERROR;
+	}
+#endif
 	serial_init(1);
 	if ((rc = reg_probe()) < 0)
 		goto ERROR;
@@ -503,6 +530,7 @@ static int bt_mod_init(void)
 		goto ERROR;
 	if (is_kern_pid_by_hook(mode))
 		enable = cpu_online_map;
+	rdtscll(start_clocks);
 	printk("bt_mod loaded(%d)\n", mode);
 	return rc;
 ERROR:
@@ -536,18 +564,20 @@ MODULE_PARM_DESC(subbuf_num, "relayfs buffer number.");
 module_param(subbuf_sleep_threshold, int, 0);
 MODULE_PARM_DESC(subbuf_sleep_threshold,
 		 "target process sleep threshold of relayfs buffers.");
-module_param(pid_max, int, PID_MAX_DEFAULT);
+module_param(pid_max, int, 0);
 MODULE_PARM_DESC(pid_max, "max pid.");
 module_param(mode, int, 0);
 MODULE_PARM_DESC(mode, "trace mode.");
+module_param(idt_table_p, ulong, 0);
 module_param(irq_desc_p, ulong, 0);
 module_param(no_irq_type_p, ulong, 0);
 module_param(switch_to_addr, ulong, 0);
 module_param(switch_to_size, int, 0);
+module_param(kallsyms_lookup_name_addr, ulong, 0);
 
 module_param(chk_sctime, int, 0);
 
-MODULE_AUTHOR("Yumiko Sugita <sugita@sdl.hitachi.co.jp>,\n" \
+MODULE_AUTHOR("Yumiko Sugita <yumiko.sugita.yf@hitachi.com>,\n" \
 	      "\t\tSatoshi Fujiwara <sa-fuji@sdl.hitachi.co.jp>");
 MODULE_DESCRIPTION("Hardware BTS facility controller");
 MODULE_LICENSE("GPL");
