@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_relfs.c - branch trace module (proc file system interface)            */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -44,8 +44,10 @@ static struct proc_dir_entry *proc_filter_syscall;
 static struct proc_dir_entry *proc_exec_syscall;
 static struct proc_dir_entry *proc_btsbuf_sz;
 static struct proc_dir_entry *proc_bts_mrecs;
+static struct proc_dir_entry *proc_irq_addrs;
 extern DEFINE_PER_CPU(struct info_per_cpu, bt_info_per_cpu);
 extern cpumask_t enable;
+extern unsigned long idt_table_p;
 extern unsigned long *pid_tbl;
 extern unsigned long *syscall_pid_tbl;
 extern unsigned long syscall_exe_tbl[];
@@ -64,6 +66,39 @@ void add_pid_tbl(pid_t, unsigned long*);
 void remove_pid_tbl(pid_t, unsigned long*);
 void add_syscall_tbl(int, unsigned long*);
 
+static int proc_irq_addrs_read(char *buf, char **start, off_t off,
+			       int count, int *eof, void *data)
+{
+	int i, len, max = 10;
+	off_t loff;
+	char lbuf[max + 1], *p;
+	struct desc_struct *idt_table = (struct desc_struct*)idt_table_p;
+
+	loff = 0;
+	*start = p = buf;
+	
+	for (i = 0; i < 256 + 1; i++) {
+		if (i == 256) {
+			unsigned long low, high;
+
+			rdmsr(MSR_IA32_SYSENTER_EIP, low, high);
+			len = snprintf(lbuf, max, "%08lx\n", low);
+		} else
+			len = snprintf(lbuf, max, "%08lx\n",
+				       (idt_table[i].a & 0x0000ffff) |
+				       (idt_table[i].b & 0xffff0000));
+		if (len < 0 || len >= max)
+			return -EFAULT;
+		if (loff >= off && loff + len <= off + count) {
+			memcpy(p, lbuf, len);
+			p += len;
+		} else if (loff + len > off + count)
+			break;
+		loff += len;
+	}
+	return p - buf;
+}
+
 static int proc_read_bitmap(char *buf, char **start, off_t off,
 			    int count, int *eof, void *data,
 			    unsigned long *tbl, int n_max)
@@ -373,6 +408,11 @@ int proc_init(void)
 		return -ENOMEM;
 	proc_bts_mrecs->read_proc = proc_bts_mrecs_read;
 
+	proc_irq_addrs = create_proc_entry("irq_addrs", 0400, proc_btrax);
+	if (!proc_irq_addrs)
+		return -ENOMEM;
+	proc_irq_addrs->read_proc = proc_irq_addrs_read;
+
 	for_each_online_cpu(i) {
 		info = &per_cpu(bt_info_per_cpu, i);
 		snprintf(tmp, sizeof(tmp), "cpu%d", i);
@@ -416,6 +456,8 @@ void proc_cleanup(void)
 		remove_proc_entry(proc_btsbuf_sz->name, proc_btrax);
 	if (proc_bts_mrecs)
 		remove_proc_entry(proc_bts_mrecs->name, proc_btrax);
+	if (proc_irq_addrs)
+		remove_proc_entry(proc_irq_addrs->name, proc_btrax);
 	if (proc_btrax)
 		remove_proc_entry(proc_btrax->name, &proc_root);
 }
