@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_relfs.c - branch trace module (relayfs interface)                     */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -59,6 +59,7 @@ extern size_t subbuf_size;
 extern size_t subbuf_num;
 extern size_t subbuf_sleep_threshold;
 extern int mode;
+extern unsigned long long start_clocks;
 
 #ifdef DEBUG
 unsigned long wcnt[NR_CPUS];
@@ -100,12 +101,10 @@ void write_pid_record(pid_t pid, char *comm)
 {
 	struct pid_record rec;
 	struct comm_record comm_rec;
-	struct timeval tv;
 
-	do_gettimeofday(&tv);
+	rdtscll(rec.clocks);
+	rec.clocks = (rec.clocks - start_clocks) | BT_FLAG_PID_LL;
 	rec.pid = pid;
-	rec.tv_sec = tv.tv_sec;
-	rec.tv_usec = tv.tv_usec | BT_FLAG_PID;
 	relfs_relay_write(bt_channel, &rec, sizeof(rec));
 
 	memcpy(comm_rec.comm, comm, BT_COMM_LEN);
@@ -383,9 +382,12 @@ int relfs_init(void)
 #if RELAYFS_CHANNEL_VERSION < 5
 	bt_channel = relay_open(relfs_basename, dir, subbuf_size, subbuf_num,
 				!is_start(mode), &relayfs_callbacks);
-#else
+#elif (RELAYFS_CHANNEL_VERSION == 5) || (RELAYFS_CHANNEL_VERSION == 6)
 	bt_channel = relay_open(relfs_basename, dir, subbuf_size, subbuf_num,
 				&relayfs_callbacks);
+#else
+	bt_channel = relay_open(relfs_basename, dir, subbuf_size, subbuf_num,
+				&relayfs_callbacks, NULL);
 #endif
 	if (bt_channel == NULL) {
 		printk("%s: cannot open relayfs channel\n", MOD_NAME);
