@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_execpath.c - execution path display program                           */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -29,8 +29,8 @@
 #include "bt_ar_parse.h"
 #include "bt_utils.h"
 
-#define BT_EXECPATH_VER	BT_VER
-#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-2007"
+#define BT_EXECPATH_VER	VERSION
+#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-" RELEASE_YEAR
 
 struct proc_each_rec_data {
 	struct addr_range	*range;
@@ -48,7 +48,7 @@ struct proc_each_range_data {
 
 int output_summary = 0;
 int verbose = 0;
-struct r2n_info r2n_info = { NULL, 128, 0 };
+struct r2n_info r2n_info;
 
 /*----------------------------------------------------------------------------
  *  get function name and mnemonic
@@ -60,7 +60,7 @@ int print_exec_line(unsigned long addr, struct range_to_name *r2n)
 	size_t offset;
 	const char *src_name, *func_name;
 	struct bfd_if *bi = &r2n->bi;
-	bfd_vma baddr;
+	unsigned long baddr;
 
 	addr -= r2n->offset;
 	printf("%s\t0x%08lx\t", r2n->basename, addr);
@@ -179,56 +179,6 @@ int proc_each_record_for_normal(struct bt_record *__p, off_t i_rec, void *tmp)
  */
 #define PRINTABLE_NEST_MAX	20
 #define PRINTABLE_NEST_STEP	20
-#define MAX_NEST	2048
-
-static int min_nest;
-static int nest;
-
-#define RET_CHK
-#ifdef RET_CHK
-struct ret_chk {
-	int		type;
-	unsigned long	addr;
-} ret_chk[MAX_NEST];
-
-static struct ret_chk *p_ret_chk_last = ret_chk;
-
-void dump_ret_chk(int n)
-{
-	struct ret_chk *p;
-
-	for (p = ret_chk + n; p < p_ret_chk_last; p++)
-		printf("%04d:0x%08lx(%d)\n",
-		       n, ret_chk[n].addr, ret_chk[n].type);
-}
-
-void set_expect_ret_addr(int type, unsigned long addr)
-{
-	if (p_ret_chk_last - ret_chk >= MAX_NEST) {
-		printf("WARN: MAX_NEST is too small.\n");
-		return;
-	}
-	p_ret_chk_last->type = type;
-	p_ret_chk_last->addr = addr;
-	p_ret_chk_last++;
-	nest++;
-}
-
-int get_nest_from_ret_addr(unsigned long addr)
-{
-	struct ret_chk *p;
-
-	for (p = p_ret_chk_last; p >= ret_chk; p--) {
-		if (p->addr == addr) {
-			p_ret_chk_last = p;
-			return p - ret_chk;
-		}
-	}
-	if (p_ret_chk_last > ret_chk)
-		p_ret_chk_last--;
-	return p_ret_chk_last - ret_chk;
-}
-#endif
 
 struct summary {
 	int type;
@@ -333,74 +283,12 @@ struct chk_repeat_funcs chk_repeat_funcs_S =
 	__print_end_repeat_S,
 };
 
-int proc_enter(int type, int nest_chk, struct bt_record *p, unsigned long ret)
-{
-	struct summary *s;
-
-	switch (type) {
-	case BTYPE_JMP:
-	case BTYPE_CALL:
-	case BTYPE_INT:
-	case BTYPE_BREAK:
-		/*
-		printf("ENTER: %d, from:%08lx, to:%08lx, ret:%08lx\n",
-		       type, p->from, p->to, ret);
-		       */
-		if (!nest_chk) {
-			s = xcalloc(1, sizeof(struct summary));
-			s->type = type;
-			s->nest = nest;
-			s->from = p->from;
-			s->to = p->to;
-			if (chk_repeat_each((const void*)s) < 0)
-				return -1;
-		}
-		if (type != BTYPE_JMP) {
-#ifdef RET_CHK
-			/* !!! FIXME: Function call by jmp is not checked */
-			set_expect_ret_addr(type, ret);
-			/* for DEBUG
-			if (!nest_chk)
-				printf("enter(%d):0x%08lx(%d)\n",
-				       nest, ret, type);
-			*/
-#else
-			nest++;
-#endif
-		}
-		break;
-	}
-	return 0;
-}
-
-void proc_leave(int type, int nest_chk, struct bt_record *p)
-{
-	if (type != BTYPE_CALL && type != BTYPE_INT && type != BTYPE_BREAK)
-		return;
-	/*
-	printf("LEAVE: %d, from:%08lx, to:%08lx\n",
-	       type, p->from, p->to);
-	       */
-#ifdef RET_CHK
-	/* for DEBUG
-	if (!nest_chk)
-		printf("leave(%d):0x%08lx(%d)\n", nest, p->to, type);
-	*/
-	nest = get_nest_from_ret_addr(p->to);
-#else
-	nest--;
-#endif
-	if (nest_chk && nest < min_nest)
-		min_nest = nest;
-}
-
 int proc_each_record_for_summary(struct bt_record *__p, off_t i_rec, void *tmp)
 {
 	struct bt_record p = *__p;
 	struct proc_each_rec_data *data = (struct proc_each_rec_data*)tmp;
-	struct range_to_name *r2n, *r2n_to;
-	int nest_chk = data->nest_chk, rc, type;
-	struct path *p_from;
+	int nest_chk = data->nest_chk, rc, inst_type, type, nest;
+	struct summary *s;
 
 	if (!nest_chk && is_warn_record(&p)) {
 		struct warn_record *warn = (struct warn_record*)&p;
@@ -409,72 +297,38 @@ int proc_each_record_for_summary(struct bt_record *__p, off_t i_rec, void *tmp)
 	}
 	if (is_pid_record(&p) || is_comm_record(&p))
 		return 0;
-	r2n_to = addr_to_r2n(&r2n_info, p.to);
-	rc = chk_fix_from_cache(&r2n_info, &p.from, p.to, &r2n, &p_from, NULL);
+	rc = chk_enter_leave(&r2n_info, &p, nest_chk, &inst_type, &type, &nest,
+			     NULL, NULL, NULL, NULL);
 	if (rc < 0)
-		return rc;
-	/*
-	if (!is_addr_range_match(p.from, p.to, data->range))
+		return -1;
+	if (!(rc == 1 || rc == 2))
 		return 0;
-		*/
-	if (rc == 0) {
-		if (!r2n_to)
-			return 0;
-		/* iret */
-		proc_leave(BTYPE_INT, nest_chk, &p);
-		return 0;
-	}
-	if (!r2n_to) {
-		/* int */
-		return proc_enter(BTYPE_INT, nest_chk, &p, p.from);
-	}
-	type = p_from->type;
-	if (p_from->base != p.from)
-		type = BTYPE_OTHER;
-	/* for DEBUG
-	if (!nest_chk)
-		printf("branch info(%d) from:%08lx branch:%08lx next:%08lx\n",
-		       type, p.from, p_from->jmp, p_from->next);
-		       */
+	/* normal jump or enter */
 	switch (type) {
-	case BTYPE_BRANCH:
 	case BTYPE_JMP:
 	case BTYPE_CALL:
-		if (p_from->jmp == UNKNOWN_BADDR || p_from->jmp == p.to) {
-			/* branch, jmp, call */
-			return proc_enter(type, nest_chk, &p,
-					  p_from->next);
-		} else {
-			/* int */
-			return proc_enter(BTYPE_INT, nest_chk, &p, p.from);
-		}
 	case BTYPE_INT:
 	case BTYPE_BREAK:
-	case BTYPE_OTHER:
-		/* int */
-		return proc_enter(BTYPE_INT, nest_chk, &p, p.from);
-	case BTYPE_RET:
-		proc_leave(BTYPE_CALL, nest_chk, &p);
-		break;
-	case BTYPE_IRET:
-		proc_leave(BTYPE_INT, nest_chk, &p);
-		break;
+		if (!nest_chk) {
+			s = xcalloc(1, sizeof(struct summary));
+			s->type = type;
+			s->nest = nest;
+			s->from = p.from;
+			s->to = p.to;
+			if (chk_repeat_each((const void*)s) < 0)
+				return -1;
+		}
 	}
 	return 0;
 }
 
 int chk_nest(int fd, off_t from, off_t to, struct proc_each_rec_data *data)
 {
-	nest = 0;
+	chk_nest_initialize();
 	if (for_each_block_record(fd, from, to,
 				  proc_each_record_for_summary, data) < 0)
 		return -1;
-	////printf("checked nest1: %d\n", nest);
-	nest = 0 - min_nest;
-#ifdef RET_CHK
-	p_ret_chk_last = ret_chk;
-#endif
-	////printf("checked nest2: %d\n", nest);
+	chk_nest_initialize();
 	return 0;
 }
 
@@ -576,8 +430,10 @@ int proc_block_records(int fd, off_t from, off_t to)
 	data.range = get_pid_addr_range(ALL_PID);
 	if (output_summary) {
 		data.nest_chk = 1;
+		/*
 		if (chk_nest(fd, from, to, &data) < 0)
 			return -1;
+			*/
 		data.nest_chk = 0;
 		chk_repeat_start(&chk_repeat_funcs_S);
 		if (for_each_block_record(fd, from, to,
@@ -637,7 +493,7 @@ int proc_logfile(char *logfile)
 			       PM_WITH_PT, PM_VERBOSE) <0)
 			goto EXIT;
 	}
-	if (parse_modules(&r2n_info, dir, NULL, PM_WITH_PT, PM_VERBOSE) < 0)
+	if (parse_modules(&r2n_info, dir, PM_WITH_PT, PM_VERBOSE) < 0)
 		goto EXIT;
 	r2n_info.from_is_next = chk_from_is_next(dir);
 
@@ -729,8 +585,8 @@ int main(int argc, char *argv[])
 
 	if (alloc_pid_range(ALL_PID) < 0)
 		err_exit();
-	while ((c = getopt_long(argc, argv, "sa:d:S:F:B:A:T:f:Wv", long_options,
-				&opt_index)) != -1) {
+	while ((c = getopt_long(argc, argv, "sa:d:S:F:B:A:T:f:WvD:",
+				long_options, &opt_index)) != -1) {
 		switch (c) {
 		case 0:
 			switch (opt_index) {
@@ -798,6 +654,9 @@ int main(int argc, char *argv[])
 		case 'v':
 			verbose = 1;
 			break;
+		case 'D':
+			set_elf_path_prefix(optarg);
+			break;
 		default:
 			usage();
 			err_exit();
@@ -810,8 +669,10 @@ int main(int argc, char *argv[])
 		usage();
 		err_exit();
 	}
+	set_uname(&r2n_info, logfile);
 	if (search_warn)
 		after = 0;
+	r2n_info.r2n_max = 16;
 	r2n_info.all_r2n = xcalloc(r2n_info.r2n_max,
 				   sizeof(struct range_to_name*));
 	if (proc_logfile(logfile) < 0)
