@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_read.c - relayfs read program                                         */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -34,13 +34,13 @@
 #include <fcntl.h>
 #include <dirent.h>
 #include <errno.h>
-#include <linux/threads.h>
 #include <sys/mman.h>
 #include <getopt.h>
-#include "bt.h"
+#include <libiberty.h>
+#include "bt_utils.h"
 
-#define	BT_READ_VER	BT_VER
-#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-2006"
+#define	BT_READ_VER	VERSION
+#define	COPYRIGHT	"Copyright (c) Hitachi, Ltd. 2005-" RELEASE_YEAR
 
 static int verbose;
 static int do_stdio;
@@ -256,64 +256,53 @@ struct in_out_pair {
 	char	basename[NAME_MAX+1];
 };
 
-static struct in_out_pair pair_tbl[NR_CPUS];
+static int nr_cpus;
+static struct in_out_pair *pair_tbl;
+
+static int __open_cpu_files(int cpu, char *cpu_path, size_t size, void *dt)
+{
+	char *out_dir = (char*)dt;
+	struct in_out_pair *p;
+	char path[PATH_MAX];
+
+	p = &pair_tbl[cpu];
+	sprintf(p->basename, "cpu%d", cpu);
+	if ((p->in = open(cpu_path, O_RDONLY | O_NONBLOCK)) < 0) {
+		fprintf(stderr, "%s open failed.(%s)\n", cpu_path,
+			strerror(errno));
+		return -1;
+	}
+	snprintf(path, PATH_MAX, "%s/%s", out_dir, p->basename);
+	if ((p->out = open(path, O_CREAT|O_EXCL|O_RDWR, S_IRUSR|S_IWUSR)) < 0) {
+		fprintf(stderr, "%s open failed.(%s)\n", path,
+			strerror(errno));
+		return -1;
+	}
+	return 0;
+}
 
 int do_read(char *in_dir, char *out_dir)
 {
-	DIR *dir;
-	struct dirent* d;
-	int max = 128;
-	char path[max];
-	struct stat st;
-	struct in_out_pair *p, *p_max;
-	int rc;
+	int cpu;
+	struct in_out_pair *p;
+
+	nr_cpus = count_cpu_files(in_dir);
+	if (nr_cpus < 0)
+		return -1;
+	pair_tbl = xcalloc(nr_cpus, sizeof(struct in_out_pair));
 
 	if (do_stdio)
 		return onefile_copy(0, 1);
 
 	/* first, check in/out files and get file descriptors */
-	if ((dir = opendir(in_dir)) == NULL)
+	if (proc_cpu_files(in_dir, __open_cpu_files, out_dir) < 0)
 		return -1;
-	p = p_max = pair_tbl;
-	rc = 0;
-	while ((d = readdir(dir)) != NULL) {
-		if (strcmp(".", d->d_name) == 0 || strcmp("..", d->d_name) == 0)
-			continue;
-		snprintf(path, max, "%s/%s", in_dir, d->d_name);
-		if (lstat(path, &st) < 0) {
-			fprintf(stderr, "%s lstat failed.(%s)\n", path,
-				strerror(errno));
-			rc = -1;
-			break;
-		}
-		if (S_ISDIR(st.st_mode))
-			continue;
-
-		/* process cpuX file */
-		strcpy(p->basename, d->d_name);
-		if ((p->in = open(path, O_RDONLY | O_NONBLOCK)) < 0) {
-			fprintf(stderr, "%s open failed.(%s)\n", path,
-				strerror(errno));
-			rc = -1;
-			break;
-		}
-		snprintf(path, max, "%s/%s", out_dir, d->d_name);
-		if ((p->out = open(path, O_CREAT|O_EXCL|O_RDWR,
-				   S_IRUSR|S_IWUSR)) < 0) {
-			fprintf(stderr, "%s open failed.(%s)\n", path,
-				strerror(errno));
-			rc = -1;
-			break;
-		}
-		p++;
-	}
-	closedir(dir);
-	if (rc < 0)
-		return rc;
 
 	/* fork each read processes */
-	p_max = p;
-	for (p = pair_tbl; p < p_max; p++) {
+	for (cpu = 0; cpu < nr_cpus; cpu++) {
+		p = &pair_tbl[cpu];
+		if (!p->in || !p->out)
+			continue;
 		switch ((p->pid = fork())) {
 		case -1:
 			fprintf(stderr, "fork failed.(%s)\n", strerror(errno));
@@ -330,7 +319,8 @@ int do_read(char *in_dir, char *out_dir)
 			break;
 		}
 	}
-	for (p = pair_tbl; p < p_max; p++) {
+	for (cpu = 0; cpu < nr_cpus; cpu++) {
+		p = &pair_tbl[cpu];
 		if (p->pid > 0)
 			waitpid(p->pid, NULL, 0);
 	}
