@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bfd_if.c - BFD library interface                                         */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -25,6 +25,7 @@
 /*****************************************************************************/
 
 #include <elf.h>
+#include <argp.h>
 #include "bt.h"
 #include "bfd_if.h"
 
@@ -45,8 +46,7 @@ static void dump_symbols(const char *header, asymbol **symbols, long cnt)
 	for (i = 0; i < cnt; i++) {
 		sym = symbols[i];
 		sect = sym->section;
-		printf("\tSYM ");
-		printf_bfd_vma(bfd_asymbol_value(sym));
+		printf("\tSYM 0x%08lx", (unsigned long)bfd_asymbol_value(sym));
 		printf(" <%s>%08x\t<%s>%08x:%d\n",
 		       sym->name, sym->flags, sect->name, sect->flags,
 		       sect->index);
@@ -62,8 +62,7 @@ static void dump_synsymbols(const char *header, asymbol *symbols, long cnt)
 	for (i = 0; i < cnt; i++) {
 		sym = &symbols[i];
 		sect = sym->section;
-		printf("\tSYM ");
-		printf_bfd_vma(bfd_asymbol_value(sym));
+		printf("\tSYM 0x%08lx", (unsigned long)bfd_asymbol_value(sym));
 		printf(" <%s>%08x\t<%s>%08x:%d\n",
 		       sym->name, sym->flags, sect->name, sect->flags,
 		       sect->index);
@@ -78,10 +77,10 @@ static void dump_relocs(struct bfd_if *bi)
 		r = bi->p_relocs[i];
 		if (r->howto->type != RELOC_TYPE_PC32)
 			continue;
-		printf("RELOC=> CODE-OFFSET:");
-		printf_bfd_vma(r->address);
-		printf(" SYM-ADDR:");
-		printf_bfd_vma((*r->sym_ptr_ptr)->value);
+		printf("RELOC=> CODE-OFFSET:0x%08lx",
+		       (unsigned long)r->address);
+		printf(" SYM-ADDR:0x%08lx",
+		       (unsigned long)(*r->sym_ptr_ptr)->value);
 		printf(" <%s>", (*r->sym_ptr_ptr)->name);
 		printf("\n");
 	}
@@ -98,14 +97,15 @@ static int is_code_sect(struct bfd_if *bi, asection *sect)
 	return 0;
 }
 
-static asection* get_sect_has_addr(struct bfd_if *bi, bfd_vma addr)
+static asection* get_sect_has_addr(struct bfd_if *bi, unsigned long addr)
 {
 	asection *s;
 
 	for (s = bi->abfd->sections; s; s = s->next) {
 		if (!is_code_sect(bi, s))
 			continue;
-		if (addr >= s->vma && addr < s->vma + bfd_get_section_size(s))
+		if (addr >= (unsigned long)s->vma &&
+		    addr < (unsigned long)(s->vma + bfd_get_section_size(s)))
 			return s;
 	}
 	return NULL;
@@ -125,10 +125,9 @@ static void test_print_reloc(struct bfd_if *bi, arelent *r)
 
 	if (!reloc_test)
 		return;
-	s = get_sect_has_addr(bi, r->address);
+	s = get_sect_has_addr(bi, (unsigned long)r->address);
 	//printf("Resolved reloc: ");
-	printf("%s ", s->name);
-	printf_bfd_vma(r->address - s->vma);
+	printf("%s 0x%08lx", s->name, (unsigned long)(r->address - s->vma));
 	printf(" %s\n", (*r->sym_ptr_ptr)->name);
 }
 #else
@@ -140,7 +139,7 @@ static int f_cmp_reloc_to_addr(const void *p1, const void *p2)
 	bfd_vma addr;
 	arelent *r;
 
-	addr = *(bfd_vma*)p1;
+	addr = (bfd_vma)(*(unsigned long*)p1);
 	r = *(arelent**)p2;
 	if (addr < r->address)
 		return -1;
@@ -149,7 +148,7 @@ static int f_cmp_reloc_to_addr(const void *p1, const void *p2)
 	return 0;
 }
 
-static int is_reloc_branch(struct bfd_if *bi, bfd_vma next)
+static int is_reloc_branch(struct bfd_if *bi, unsigned long next)
 {
 	arelent **p;
 
@@ -161,8 +160,10 @@ static int is_reloc_branch(struct bfd_if *bi, bfd_vma next)
 	return (p != NULL);
 }
 
-static bfd_vma get_relocated_baddr(struct bfd_if *bi, bfd_vma branch,
-				   bfd_vma next, bfd_vma offset)
+static unsigned long get_relocated_baddr(struct bfd_if *bi,
+					 unsigned long branch,
+					 unsigned long next,
+					 unsigned long offset)
 {
 	arelent **p;
 	asymbol *sym;
@@ -176,13 +177,13 @@ static bfd_vma get_relocated_baddr(struct bfd_if *bi, bfd_vma branch,
 		return branch + offset;
 	test_print_reloc(bi, *p);
 	sym = *(*p)->sym_ptr_ptr;
-	branch = branch - next + bfd_asymbol_value(sym);
+	branch = branch - next + (unsigned long)bfd_asymbol_value(sym);
 	if (sym->flags & BSF_FUNCTION) {
 		/* symbol defined same section of this module */
 		return branch + offset;
 	} else if (sym->flags & BSF_SECTION_SYM) {
 		/* symbol defined other section of this module */
-		offset += sym->section->vma;
+		offset += (unsigned long)sym->section->vma;
 		return branch + offset;
 	} else {
 		/* symbol defined outside the module */
@@ -190,14 +191,15 @@ static bfd_vma get_relocated_baddr(struct bfd_if *bi, bfd_vma branch,
 	}
 }
 
-static int prepare_print_insn(struct bfd_if *bi, bfd_vma addr)
+static int prepare_print_insn(struct bfd_if *bi, unsigned long addr)
 {
 	asection *sect;
 	struct disassemble_info *info = &bi->info;
 	struct sect_cache *cache = &bi->cache;
 
 	if (cache->data) {
-		if (addr >= cache->vma && addr < cache->vma + cache->size) {
+		if (addr >= (unsigned long)cache->vma &&
+		    addr < (unsigned long)(cache->vma + cache->size)) {
 			goto CACHE_FINISH;
 		} else {
 			free(cache->data);
@@ -339,11 +341,21 @@ static void set_sect_offset(struct bfd_if *bi)
 		return;
 	sect_offset = 0;
 	for (s = bi->abfd->sections; s; s = s->next) {
-		if (!is_code_sect(bi, s))
+		/* If the section address already set, then do not add the
+		 * section offset (e.g. relocatable kernel).
+		 */
+		if (!is_code_sect(bi, s) || s->vma)
 			continue;
 		s->vma += sect_offset;
 		sect_offset += bfd_get_section_size(s);
 	}
+	/* for DEBUG
+	for (s = bi->abfd->sections; s; s = s->next) {
+		if (!is_code_sect(bi, s))
+			continue;
+		printf("code-sect:%s\t%08lx\n", s->name, s->vma);
+	}
+	*/
 	return;
 }
 
@@ -500,7 +512,7 @@ static int resolve_undsyms(struct bfd_if *bi, const char *obj_name,
 	buf[MAX_LINE_LEN] = '\0';
 	while ((p = fgets(buf, MAX_LINE_LEN, f)) != NULL) {
 		len = strlen(buf);
-		if (buf[len - 2] != ']' || buf[9] != 'U')
+		if (buf[len - 2] != ']' || (buf[9] != 'U' && buf[9] != 'u'))
 			continue;
 		num = sscanf(buf, "%lx %*c %s\t[%[^]]]", &addr, sym, mod);
 		if (num != 3 || strcmp(mod_name, mod) != 0)
@@ -623,6 +635,8 @@ EXIT:
 		free(bi->p_synsyms);
 	if (bi->p_fsyms)
 		free(bi->p_fsyms);
+	if (bi->p_relocs)
+		free(bi->p_relocs);
 	return -1;
 }
 
@@ -638,6 +652,19 @@ static void chk_has_debuginfo(struct bfd_if *bi)
 		}
 }
 
+#ifdef USE_ELFUTILS
+int init_dwarf(struct bfd_if *bi, const char *obj_name)
+{
+	int argi, n = 3;
+	char *s[] = { "", "-e", (char*)obj_name, NULL };
+
+	argp_parse(dwfl_standard_argp(), n, s, 0, &argi, &bi->dwfl);
+	if (!bi->dwfl)
+		return -1;
+	return 0;
+}
+#endif
+
 int init_bfd_if(struct bfd_if *bi, const char *obj_name, const char *kallsyms)
 {
 	struct disassemble_info *info = &bi->info;
@@ -681,7 +708,7 @@ int init_bfd_if(struct bfd_if *bi, const char *obj_name, const char *kallsyms)
 	return 0;
 }
 
-bfd_vma get_offset_addr(struct bfd_if *bi, bfd_vma begin)
+unsigned long get_offset_addr(struct bfd_if *bi, unsigned long begin)
 {
 	if (bfd_get_file_flags(bi->abfd) & HAS_RELOC)
 		return begin;
@@ -690,7 +717,8 @@ bfd_vma get_offset_addr(struct bfd_if *bi, bfd_vma begin)
 
 /* This function is specified for kernel. It doesn't calculate section offset.
  */
-void get_kernel_min_max_addr(struct bfd_if *bi, bfd_vma *min, bfd_vma *max)
+void get_kernel_min_max_addr(struct bfd_if *bi,
+			     unsigned long *min, unsigned long *max)
 {
 	bfd_vma sect_max;
 	asection *s;
@@ -700,17 +728,17 @@ void get_kernel_min_max_addr(struct bfd_if *bi, bfd_vma *min, bfd_vma *max)
 	for (s = bi->abfd->sections; s; s = s->next) {
 		if (!is_code_sect(bi, s))
 			continue;
-		if (s->vma < *min)
-			*min = s->vma;
+		if ((unsigned long)s->vma < *min)
+			*min = (unsigned long)s->vma;
 		sect_max = s->vma + bfd_get_section_size(s) - 1;
-		if (sect_max > *max)
-			*max = sect_max;
+		if ((unsigned long)sect_max > *max)
+			*max = (unsigned long)sect_max;
 	}
 	return;
 }
 
-void remove_useless_fsyms(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
-			  bfd_vma offset)
+void remove_useless_fsyms(struct bfd_if *bi, unsigned long begin,
+			  unsigned long end, unsigned long offset)
 {
 	long i, cnt;
 	asymbol *sym, **symbols = bi->p_fsyms;
@@ -727,7 +755,8 @@ void remove_useless_fsyms(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
 		if (val < sect->vma ||
 		    val >= sect->vma + bfd_get_section_size(sect))
 			continue;
-		if (val + offset < begin || val + offset > end)
+		if ((unsigned long)val + offset < begin ||
+		    (unsigned long)val + offset > end)
 			continue;
 
 		if (val != prev_val) {
@@ -743,7 +772,7 @@ void remove_useless_fsyms(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
  *-----------------------------------------------------------------------------
  */
 static int btype;
-static bfd_vma baddr;
+static unsigned long baddr;
 
 #define PRINT_BUF_MAX	128
 static char print_buf[PRINT_BUF_MAX + 1];
@@ -783,8 +812,7 @@ static int chk_btype_dummy_fprintf(FILE *stream, const char *fmt, ...)
 		} else if (strncmp(print_buf, "int", 3) == 0 ||
 			   strncmp(print_buf, "sysenter", 8) == 0) {
 			btype = BTYPE_INT;
-		} else if (strncmp(print_buf, "hlt", 3) == 0 ||
-			   strncmp(print_buf, "ud2", 3) == 0) {
+		} else if (strncmp(print_buf, "ud2", 3) == 0) {
 			btype = BTYPE_BREAK;
 		} else
 			btype = BTYPE_OTHER;
@@ -812,8 +840,9 @@ static int chk_btype_dummy_fprintf(FILE *stream, const char *fmt, ...)
 	return rc;
 }
 
-static int __get_branch_info(struct bfd_if *bi, bfd_vma addr, int *type,
-			     bfd_vma *addr_next, bfd_vma *addr_branch)
+static int __get_branch_info(struct bfd_if *bi, unsigned long addr, int *type,
+			     unsigned long *addr_next,
+			     unsigned long *addr_branch)
 {
 	int bytes;
 	struct disassemble_info *info = &bi->info;
@@ -828,7 +857,7 @@ static int __get_branch_info(struct bfd_if *bi, bfd_vma addr, int *type,
 	info->fprintf_func = (fprintf_ftype)chk_btype_dummy_fprintf;
 	//bytes = print_insn_i386_intel(addr, info);
 	//bytes = print_insn_i386(addr, info);
-	bytes = print_insn_i386_att(addr, info);
+	bytes = print_insn_i386_att((bfd_vma)addr, info);
 	if (bytes < 0) {
 		last_error |= CHK_BTYPE_E_INTERNAL;
 		goto EXIT;
@@ -848,8 +877,9 @@ EXIT:
 	return (last_error ? -last_error : 0);
 }
 
-int get_branch_info(struct bfd_if *bi, bfd_vma addr, int *type,
-		    bfd_vma *next, bfd_vma *branch, bfd_vma offset)
+int get_branch_info(struct bfd_if *bi, unsigned long addr, int *type,
+		    unsigned long *next, unsigned long *branch,
+		    unsigned long offset)
 {
 	if (__get_branch_info(bi, addr - offset, type, next, branch) < 0)
 		return -1;
@@ -867,9 +897,10 @@ typedef struct {
 	struct bfd_if		*bi;
 	node			*pt;
 	node			*node_addrs;
+	int			skip_ud2_srcinfo;
 } pack_bi_pt;
 
-static int ____f_chk_range(struct path *p, bfd_vma addr)
+static int ____f_chk_range(struct path *p, unsigned long addr)
 {
 	if (p->cnt < 0) {
 		if (addr < p->addr)
@@ -887,21 +918,21 @@ static int ____f_chk_range(struct path *p, bfd_vma addr)
 
 static int __f_chk_range(void *__target, void *__dt)
 {
-	bfd_vma addr = *((bfd_vma*)__target);
+	unsigned long addr = *((unsigned long*)__target);
 	struct path *p = (struct path*)__dt;
 
 	return ____f_chk_range(p, addr);
 }
 
-static struct path* __find_path_from_addr(node *tree, bfd_vma addr)
+static struct path* __find_path_from_addr(node *tree, unsigned long addr)
 {
 	return (struct path*)search_tree(&addr, tree, __f_chk_range);
 }
 
 static int f_chk_addr(void *__target, void *__dt)
 {
-	bfd_vma a1 = *((bfd_vma*)__target);
-	bfd_vma a2 = *((bfd_vma*)__dt);
+	unsigned long a1 = *((unsigned long*)__target);
+	unsigned long a2 = *((unsigned long*)__dt);
 
 	if (a1 < a2)
 		return -1;
@@ -911,25 +942,106 @@ static int f_chk_addr(void *__target, void *__dt)
 }
 
 static int add_node_addrs(struct bfd_if *bi, node *pt, node **node_addrs,
-			  bfd_vma addr)
+			  unsigned long addr)
 {
-	bfd_vma *p;
+	unsigned long *p;
 
 	if (addr == UNKNOWN_BADDR || __find_path_from_addr(pt,addr)
 	    ||search_tree(&addr, *node_addrs, f_chk_addr))
 		return 0;
 
 #ifdef DBG_CHK_PATH
-	printf("ADD:");
-	printf_bfd_vma(addr);
-	printf("\n");
+	printf("ADD:0x%08lx\n", addr);
 #endif
-	p = xmalloc(sizeof(bfd_vma));
+	p = xmalloc(sizeof(unsigned long));
 	*p = addr;
 	*node_addrs = insert_tree(p, *node_addrs, f_chk_addr);
 	return *node_addrs == NULL ? -1 : 0;
 }
 
+/*
+ * Check ud2 instruction from 'begin' to 'end', and return the possibility
+ * of having source information.
+ * If no ud2 instruction found, it return -1.
+ */
+int __chk_ud2_src_info(struct bfd_if *bi, unsigned long begin,unsigned long end)
+{
+	unsigned long addr, next, branch;
+	int rc, type;
+	struct sect_cache *cache = &bi->cache;
+
+	for (addr = begin; addr < end; addr = next) {
+		rc = __get_branch_info(bi, addr, &type, &next, &branch);
+		if (rc < 0) {
+			fprintf(stderr, "__get_branch_info failed at 0x%08lx",
+				addr);
+			fprintf(stderr, ".(%d)\n", rc);
+			return -1;
+		}
+		/* check ud2 instruction */
+		if (type != BTYPE_BREAK ||
+		    cache->data[addr - (unsigned long)cache->vma] != 0x0f ||
+		    cache->data[addr - (unsigned long)cache->vma + 1] != 0x0b)
+			continue;
+		/* Caution:
+		 * At this point, we have to finish checking this function
+		 * because we don't know whether we should skip source
+		 * information right now.
+		 * So, we continue checking from next function's top address.
+		 */
+		/* ud2 instruction (2bytes) and src info (6bytes) */
+		addr += 2 + 5;
+		if (addr < (unsigned long)cache->vma ||
+		    addr >= (unsigned long)(cache->vma + cache->size))
+			return -1;
+		return (cache->data[addr - (unsigned long)cache->vma]
+				== (addr >> 24));
+	}
+	return -1;
+}
+
+#define MAX_CHK_UD2	10
+#define MAX_CHK_FUNC	100
+int chk_kernel_has_ud2_src_info(struct bfd_if *bi)
+{
+	int rc, i, cnt, func_cnt;
+	asymbol *sym, *sym2;
+	asection *s;
+	unsigned long begin, end, sect_end;
+
+	cnt = func_cnt = 0;
+	for (i = 0; i + 1 < bi->n_fsyms && func_cnt < MAX_CHK_FUNC; i++) {
+		sym = bi->p_fsyms[i];
+		sym2 = bi->p_fsyms[i+1];
+		if (!(sym->flags & BSF_FUNCTION))
+			continue;
+		if (!cnt)
+			func_cnt++;
+		/* get 'begin' and 'end' addresses */
+		begin = (unsigned long)bfd_asymbol_value(sym);
+		end = (unsigned long)bfd_asymbol_value(sym2);
+		s = get_sect_has_addr(bi, begin);
+		sect_end = (unsigned long)(s->vma + bfd_get_section_size(s));
+		if (end > sect_end)
+			end = sect_end;
+		//printf("checking <%s>:%08lx-%08lx\t", sym->name, begin, end);
+		rc = __chk_ud2_src_info(bi, begin, end);
+		switch (rc) {
+		case -1:
+			//printf("---\n");
+			continue;
+		case 0:
+			//printf("ud2 (NO src info)\n");
+			return 0;
+		case 1:
+			//printf("ud2 (WITH src info)\n");
+			if (++cnt >= MAX_CHK_UD2)
+				return 1;
+		}
+	}
+	return 0;
+}
+
 static int add_func_and_section_start_addrs(struct bfd_if *bi,
 					    node **node_addrs)
 {
@@ -942,37 +1054,36 @@ static int add_func_and_section_start_addrs(struct bfd_if *bi,
 		if (!(sym->flags & BSF_FUNCTION))
 			continue;
 		if (add_node_addrs(bi, NULL, node_addrs,
-				   bfd_asymbol_value(sym)) < 0)
+				   (unsigned long)bfd_asymbol_value(sym)) < 0)
 			return -1;
 	}
 	for (s = bi->abfd->sections; s; s = s->next) {
 		if (!is_code_sect(bi, s))
 			continue;
-		if (add_node_addrs(bi, NULL, node_addrs, s->vma) < 0)
+		if (add_node_addrs(bi, NULL, node_addrs, (unsigned long)s->vma)
+		    < 0)
 			return -1;
 	}
 	return 0;
 }
 
-static int chk_one_node_addr(struct bfd_if *bi, bfd_vma begin, bfd_vma max,
-			     node **node_addrs, node **pt)
+static int chk_one_node_addr(struct bfd_if *bi,
+			     unsigned long begin, unsigned long max,
+			     node **node_addrs, node **pt, int skip_ud2_srcinfo)
 {
-	bfd_vma addr, next, branch;
+	unsigned long addr, next, branch;
 	int rc, type;
 	struct path *p;
 
 #ifdef DBG_CHK_PATH
-	printf("CHK:");
-	printf_bfd_vma(begin);
-	printf(" max(");
-	printf_bfd_vma(max);
-	printf(")\n");
+	printf("CHK:0x%08lx", begin);
+	printf(" max(0x%08lx)\n", max);
 #endif
 	for (addr = begin; addr < max; addr = next) {
 		rc = __get_branch_info(bi, addr, &type, &next, &branch);
 		if (rc < 0) {
-			fprintf(stderr, "__get_branch_info failed at ");
-			fprintf_err_bfd_vma(addr);
+			fprintf(stderr, "__get_branch_info failed at 0x%08lx",
+				addr);
 			fprintf(stderr, ".(%d)\n", rc);
 			return rc;
 		}
@@ -989,8 +1100,7 @@ static int chk_one_node_addr(struct bfd_if *bi, bfd_vma begin, bfd_vma max,
 		/* When switch case jump, set next_cnt value to -1. */
 		p->next_cnt = type == BTYPE_JMP ? -1 : 0;
 		*pt = insert_tree(p, *pt, f_chk_addr);
-		if (type == BTYPE_BREAK)
-		////if (type == BTYPE_RET || type == BTYPE_IRET ||type == BTYPE_INT)
+		if (skip_ud2_srcinfo && type == BTYPE_BREAK)
 			break;
 		if (!is_reloc_branch(bi, next)) {
 			if ((rc = add_node_addrs(bi, *pt, node_addrs, branch))
@@ -1007,22 +1117,24 @@ static int chk_one_node_addr(struct bfd_if *bi, bfd_vma begin, bfd_vma max,
 
 static int f_chk_each_node_addr(void *__dt, void *user_data)
 {
-	bfd_vma addr = *((bfd_vma*)__dt);
+	unsigned long addr = *((unsigned long*)__dt);
 	pack_bi_pt *pack = (pack_bi_pt*)user_data;
 	node **node_addrs = &pack->node_addrs;
 	asection *s = get_sect_has_addr(pack->bi, addr);
 
 	if (!s)
 		return 0;
-	if (chk_one_node_addr(pack->bi, addr, s->vma + bfd_get_section_size(s),
-			      node_addrs, &pack->pt) < 0)
+	if (chk_one_node_addr(pack->bi, addr,
+			      (unsigned long)s->vma + bfd_get_section_size(s),
+			      node_addrs, &pack->pt, pack->skip_ud2_srcinfo)
+	    < 0)
 		return -1;
 	return 0;
 }
 
 static void f_free_each_node_addr(void *__dt)
 {
-	bfd_vma *p = (bfd_vma*)__dt;
+	unsigned long *p = (unsigned long*)__dt;
 
 	if (p)
 		free(p);
@@ -1110,7 +1222,7 @@ static int fix_all_path(struct bfd_if *bi, struct path **parray, long cnt)
 	long i;
 	struct path *p, *p_next;
 	int rc, type;
-	bfd_vma next, branch;
+	unsigned long next, branch;
 
 	for (i = 0; i < cnt; i++) {
 		p = parray[i];
@@ -1126,9 +1238,8 @@ static int fix_all_path(struct bfd_if *bi, struct path **parray, long cnt)
 			continue;
 		rc = __get_branch_info(bi, p->base, &type, &next, &branch);
 		if (rc < 0) {
-			fprintf(stderr, "fix_all_path failed at ");
-			fprintf_err_bfd_vma(p->base);
-			fprintf(stderr, ".(%d)\n", rc);
+			fprintf(stderr, "fix_all_path failed at 0x%08lx.(%d)\n",
+				p->base, rc);
 			return -1;
 		}
 		if (branch != 0 && branch != UNKNOWN_BADDR)
@@ -1144,11 +1255,11 @@ static int fix_all_path(struct bfd_if *bi, struct path **parray, long cnt)
 }
 
 static void resolve_relocs_and_offset(struct bfd_if *bi, struct path **parray,
-				      long cnt, bfd_vma offset)
+				      long cnt, unsigned long offset)
 {
 	long i;
 	struct path *p;
-	bfd_vma rc_addr;
+	unsigned long rc_addr;
 
 	for (i = 0; i < cnt; i++) {
 		p = parray[i];
@@ -1177,13 +1288,14 @@ static void resolve_relocs_and_offset(struct bfd_if *bi, struct path **parray,
  *   Second, fix the overlapping addresses.
  */
 int chk_path_tree(struct bfd_if *bi, struct path ***parray, long *cnt,
-		  bfd_vma offset)
+		  unsigned long offset, int skip_ud2_srcinfo)
 {
 	pack_bi_pt pack;
 
 	pack.bi = bi;
 	pack.pt = NULL;
 	pack.node_addrs = NULL;
+	pack.skip_ud2_srcinfo = skip_ud2_srcinfo;
 
 	/* 1st process */
 	if (add_func_and_section_start_addrs(bi, &pack.node_addrs) < 0) {
@@ -1221,6 +1333,21 @@ EXIT:
  *  path utility
  *-----------------------------------------------------------------------------
  */
+void free_path(struct path *p)
+{
+	struct unknown *uk, *next;
+
+	if (!p)
+		return;
+	if (p->jmp == UNKNOWN_BADDR) {
+		for (uk = (struct unknown*)p->jmp_cnt; uk; uk = next) {
+			next = uk->next;
+			free(uk);
+		}
+	}
+	free(p);
+}
+
 void printf_path(struct path *p)
 {
 	struct unknown *uk;
@@ -1231,7 +1358,7 @@ void printf_path(struct path *p)
 		type &= ~IS_INVALID_PATH;
 	} else
 		printf("(%ld)\t ", p->cnt);
-	printf_bfd_vma(p->addr);
+	printf("0x%08lx", p->addr);
 	switch (type) {
 	case BTYPE_BRANCH:
 		printf(" JXX  "); break;
@@ -1249,12 +1376,10 @@ void printf_path(struct path *p)
 		printf(" BRK  "); break;
 	default:
 		printf(" ---  ---------- => ----------(-) : ");
-		printf_bfd_vma(p->next);
-		printf("(%ld)\n", p->next_cnt);
+		printf("0x%08lx(%ld)\n", p->next, p->next_cnt);
 		return;
 	}
-	printf_bfd_vma(p->base);
-	printf(" => ");
+	printf("0x%08lx => ", p->base);
 	switch (type) {
 	case BTYPE_BRANCH:
 	case BTYPE_CALL:
@@ -1265,20 +1390,18 @@ void printf_path(struct path *p)
 				printf("XXXXXXXXXX(0) : ");
 			else {
 				for (; uk; uk = uk->next) {
-					printf_bfd_vma(uk->addr);
-					printf("(%ld), ", uk->cnt);
+					printf("0x%08lx(%ld), ",
+					       uk->addr, uk->cnt);
 				}
 				printf(": ");
 			}
 		} else {
-			printf_bfd_vma(p->jmp);
-			printf("(%ld) : ", p->jmp_cnt);
+			printf("0x%08lx(%ld) : ", p->jmp, p->jmp_cnt);
 		}
 		if (type == BTYPE_JMP)
 			printf("----------(-)");
 		else {
-			printf_bfd_vma(p->next);
-			printf("(%ld)", p->next_cnt);
+			printf("0x%08lx(%ld)", p->next, p->next_cnt);
 		}
 		break;
 	case BTYPE_RET:
@@ -1301,7 +1424,7 @@ void dump_path_tree(struct path **tree, long cnt)
 	/* for DEBUG
 	{
 		struct path *p_prev, *p;
-		bfd_vma addr;
+		unsigned long addr;
 
 		p_prev = tree[0];
 		printf("start: %08lx\n", p_prev->addr);
@@ -1318,20 +1441,17 @@ void dump_path_tree(struct path **tree, long cnt)
 
 static int f_chk_range(const void *p1, const void *p2)
 {
-	bfd_vma addr = *((bfd_vma*)p1);
+	unsigned long addr = *((unsigned long*)p1);
 	struct path *p = *(struct path**)p2;
 
 #ifdef DBG_CHK_PATH
-	printf("DBG:");
-	printf_bfd_vma(addr);
-	printf(":");
-	printf_bfd_vma(p->addr);
-	printf("\n");
+	printf("DBG:0x%08lx:0x%08lx\n", addr, p->addr);
 #endif
 	return ____f_chk_range(p, addr);
 }
 
-struct path* find_path_from_addr(struct path **tree, long *cnt, bfd_vma addr)
+struct path* find_path_from_addr(struct path **tree, long *cnt,
+				 unsigned long addr)
 {
 	struct path **p;
 
@@ -1348,8 +1468,51 @@ struct path* find_path_from_addr(struct path **tree, long *cnt, bfd_vma addr)
  *  other utility
  *-----------------------------------------------------------------------------
  */
-int get_source_info(struct bfd_if *bi, bfd_vma addr,
-		    const char **src_name, const char **func_name, int *lno)
+int get_source_info_binutils(struct bfd_if *bi, unsigned long addr,
+			     const char **src_name, const char **func_name,
+			     int *lno);
+#ifdef USE_ELFUTILS
+int get_source_info_elfutils(struct bfd_if *bi, unsigned long addr,
+			     const char **src_name, const char **func_name,
+			     int *lno)
+{
+	Dwfl_Module *mod = NULL;
+	Dwfl_Line *line = NULL;
+	Dwarf_Addr a = addr;
+
+	*src_name = *func_name = NULL; /* func_name is not used */
+	*lno = 0;
+
+	mod = dwfl_addrmodule(bi->dwfl, a);
+	line = dwfl_module_getsrc(mod, a);
+	if (!line ||
+	    !(*src_name = dwfl_lineinfo(line, &a, lno, NULL, NULL, NULL)))
+		return -1;
+	return 0;
+}
+
+inline int get_source_info(struct bfd_if *bi, unsigned long addr,
+			   const char **src_name, const char **func_name,
+			   int *lno)
+{
+	if (bi->dwfl)
+		return get_source_info_elfutils(bi, addr, src_name, func_name,
+						lno);
+	else
+		return get_source_info_binutils(bi, addr, src_name, func_name,
+						lno);
+}
+#else
+inline int get_source_info(struct bfd_if *bi, unsigned long addr,
+			   const char **src_name, const char **func_name,
+			   int *lno)
+{
+	return get_source_info_binutils(bi, addr, src_name, func_name, lno);
+}
+#endif
+int get_source_info_binutils(struct bfd_if *bi, unsigned long addr,
+			     const char **src_name, const char **func_name,
+			     int *lno)
 {
 	int rc;
 	asection *sect;
@@ -1366,8 +1529,8 @@ int get_source_info(struct bfd_if *bi, bfd_vma addr,
 		return -1;
 	rc = bfd_find_nearest_line(bi->abfd, sect,
 				   bi->n_syms ? bi->p_syms : bi->p_dynsyms,
-				   addr - sect->vma,
-				   src_name, func_name, lno);
+				   (bfd_vma)addr - sect->vma,
+				   src_name, func_name, (unsigned int*)lno);
 	return (rc ? 0 : -1);
 }
 
@@ -1378,9 +1541,9 @@ static void* __bsearch(const void *key, const void *base, size_t nmemb,
 		       size_t size, int (*comp)(const void*, const void*),
 		       size_t *small_index)
 {
-	size_t min, max, mid;
+	size_t min, max, mid = 0;
 	const void *p;
-	int rc;
+	int rc = 0;
 
 	min = 0;
 	max = nmemb;
@@ -1405,9 +1568,9 @@ static void* __bsearch(const void *key, const void *base, size_t nmemb,
 
 static int __comp_fsym_addr(const void *key, const void *elem)
 {
-	bfd_vma addr = *(bfd_vma*)key;
+	unsigned long addr = *(unsigned long*)key;
 	asymbol *sym = *(asymbol**)elem;
-	bfd_vma sym_addr = bfd_asymbol_value(sym);
+	unsigned long sym_addr = (unsigned long)bfd_asymbol_value(sym);
 
 	if (addr > sym_addr)
 		return 1;
@@ -1417,12 +1580,11 @@ static int __comp_fsym_addr(const void *key, const void *elem)
 }
 
 /* addr is relative */
-int addr_to_func_name_and_offset(struct bfd_if *bi, bfd_vma addr,
+int addr_to_func_name_and_offset(struct bfd_if *bi, unsigned long addr,
 				 const char **func_name, size_t *foffset)
 {
 	asymbol *sym, **p_sym;
 	size_t sym_index;
-	asection *sect;
 
 	if (addr == UNKNOWN_BADDR)
 		goto NOT_FOUND;
@@ -1434,9 +1596,8 @@ int addr_to_func_name_and_offset(struct bfd_if *bi, bfd_vma addr,
 		sym = bi->p_fsyms[sym_index];
 	} else
 		sym = *p_sym;
-	sect = sym->section;
 	*func_name = sym->name;
-	*foffset = addr - bfd_asymbol_value(sym);
+	*foffset = addr - (unsigned long)bfd_asymbol_value(sym);
 	return 0;
 
 NOT_FOUND:
@@ -1470,7 +1631,8 @@ void printf_func_name(const char *func_name, size_t offset)
 		printf("<%s>", func_name);
 }
 
-int printf_mnemonic(struct bfd_if *bi, bfd_vma addr, bfd_vma *p_baddr)
+int printf_mnemonic(struct bfd_if *bi, unsigned long addr,
+		    unsigned long *p_baddr)
 {
 	int rc;
 	char *p_tmp;
@@ -1509,7 +1671,10 @@ static bfd_vma get_sect_offset(struct bfd_if *bi, asection *sect)
 		return 0;
 	sect_offset = 0;
 	for (s = bi->abfd->sections; s; s = s->next) {
-		if (!is_code_sect(bi, s))
+		/* If the section address already set, then do not add the
+		 * section offset (e.g. relocatable kernel).
+		 */
+		if (!is_code_sect(bi, s) || s->vma)
 			continue;
 		if (s == sect)
 			return sect_offset;
@@ -1518,8 +1683,9 @@ static bfd_vma get_sect_offset(struct bfd_if *bi, asection *sect)
 	return sect_offset;
 }
 
-static int __for_each_fsyms(struct bfd_if *bi, bfd_vma offset,
-			    int (*f)(bfd_vma,bfd_vma,bfd_vma,const char*,void*),
+static int __for_each_fsyms(struct bfd_if *bi, unsigned long offset,
+			    int (*f)(unsigned long,unsigned long,unsigned long,
+				     const char*,void*),
 			    void *data)
 {
 	int rc;
@@ -1533,15 +1699,16 @@ static int __for_each_fsyms(struct bfd_if *bi, bfd_vma offset,
 		val = bfd_asymbol_value(sym);
 		sect = sym->section;
 		sect_offset = get_sect_offset(bi, sect);
-		rc = f(val, sect_offset, offset, sym->name, data);
+		rc = f((unsigned long)val, (unsigned long)sect_offset, offset,
+		       sym->name, data);
 		if (rc != 0)
 			return rc;
 	}
 	return 0;
 }
 
-int for_each_fsymbols(struct bfd_if *bi, bfd_vma offset, t_func_for_fsym f,
-		      void *data)
+int for_each_fsymbols(struct bfd_if *bi, unsigned long offset,
+		      t_func_for_fsym f, void *data)
 {
 	int rc;
 	long i;
@@ -1554,23 +1721,22 @@ int for_each_fsymbols(struct bfd_if *bi, bfd_vma offset, t_func_for_fsym f,
 		val = bfd_asymbol_value(sym);
 		sect = sym->section;
 		sect_offset = get_sect_offset(bi, sect);
-		rc = f(val + offset, sym->name, data);
+		rc = f((unsigned long)val + offset, sym->name, data);
 		if (rc != 0)
 			return rc;
 	}
 	return 0;
 }
 
-static int f_print_fsym(bfd_vma addr, bfd_vma sect_offset, bfd_vma offset,
-			const char *name, void *data)
+static int f_print_fsym(unsigned long addr, unsigned long sect_offset,
+			unsigned long offset, const char *name, void *data)
 {
-	printf_bfd_vma(addr - sect_offset);
-	printf(" %s\n", name);
+	printf("0x%08lx %s\n", addr - sect_offset, name);
 	return 0;
 }
 
-void dump_bfd_symbols(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
-		      bfd_vma offset)
+void dump_bfd_symbols(struct bfd_if *bi, unsigned long begin, unsigned long end,
+		      unsigned long offset)
 {
 	printf("------ bfd symbols ------\n");
 	__for_each_fsyms(bi, offset, f_print_fsym, NULL);
@@ -1578,11 +1744,11 @@ void dump_bfd_symbols(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
 
 struct find_fsym_data {
 	const char *name;
-	bfd_vma addr;
+	unsigned long addr;
 	int cnt;
 };
 
-static int f_find_fsym(bfd_vma addr, const char *name, void *data)
+static int f_find_fsym(unsigned long addr, const char *name, void *data)
 {
 	struct find_fsym_data *dt = (struct find_fsym_data*)data;
 
@@ -1597,8 +1763,8 @@ static int f_find_fsym(bfd_vma addr, const char *name, void *data)
 /* This function returns the number that a function name was found in.
  * In the 'addr' variable, an address of a function found first is set.
  */
-int get_addr_of_symbol(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
-		       bfd_vma offset, const char *symstr, bfd_vma *addr)
+int get_addr_of_symbol(struct bfd_if *bi, unsigned long offset,
+		       const char *symstr, unsigned long *addr)
 {
 	struct find_fsym_data dt;
 
@@ -1613,11 +1779,11 @@ int get_addr_of_symbol(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
 
 struct find_fsym_all_data {
 	const char *name;
-	bfd_vma *addrs;
+	unsigned long *addrs;
 	int cnt;
 };
 
-static int f_find_fsym_all(bfd_vma addr, const char *name, void *data)
+static int f_find_fsym_all(unsigned long addr, const char *name, void *data)
 {
 	struct find_fsym_all_data *dt = (struct find_fsym_all_data*)data;
 	int step = 16;
@@ -1627,9 +1793,10 @@ static int f_find_fsym_all(bfd_vma addr, const char *name, void *data)
 			if (dt->cnt) {
 				dt->addrs = xrealloc(dt->addrs,
 						     (dt->cnt + step) *
-						    	sizeof(bfd_vma));
+						    	sizeof(unsigned long));
 			} else {
-				dt->addrs = xmalloc(step * sizeof(bfd_vma));
+				dt->addrs = xmalloc(step *
+						    sizeof(unsigned long));
 			}
 		}
 		dt->addrs[dt->cnt++] = addr;
@@ -1637,8 +1804,8 @@ static int f_find_fsym_all(bfd_vma addr, const char *name, void *data)
 	return 0;
 }
 
-int get_addr_of_symbol_all(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
-			   bfd_vma offset, const char *symstr, bfd_vma **addrs)
+int get_addr_of_symbol_all(struct bfd_if *bi, unsigned long offset,
+			   const char *symstr, unsigned long **addrs)
 {
 	struct find_fsym_all_data dt;
 
@@ -1651,7 +1818,25 @@ int get_addr_of_symbol_all(struct bfd_if *bi, bfd_vma begin, bfd_vma end,
 	return dt.cnt;
 }
 
-int get_end_of_func(struct bfd_if *bi, bfd_vma begin, bfd_vma *end)
+int get_begin_of_func(struct bfd_if *bi, unsigned long addr,
+		      unsigned long *begin)
+{
+	asymbol *sym, **p_sym;
+	size_t sym_index;
+
+	p_sym = __bsearch(&addr, bi->p_fsyms, bi->n_fsyms, sizeof(asymbol*),
+			  __comp_fsym_addr, &sym_index);
+	if (!p_sym) {
+		if (sym_index < 0)
+			return 0;
+		sym = bi->p_fsyms[sym_index];
+	} else
+		sym = *p_sym;
+	*begin = (unsigned long)bfd_asymbol_value(sym);
+	return 1;
+}
+
+int get_end_of_func(struct bfd_if *bi, unsigned long begin, unsigned long *end)
 {
 	asymbol *sym, **p_sym;
 	long i;
@@ -1664,7 +1849,7 @@ int get_end_of_func(struct bfd_if *bi, bfd_vma begin, bfd_vma *end)
 	i = (p_sym - bi->p_fsyms);
 	if (i < bi->n_fsyms - 1) {
 		sym = bi->p_fsyms[i + 1];
-		*end = bfd_asymbol_value(sym);
+		*end = (unsigned long)bfd_asymbol_value(sym);
 	}
 	return 1;
 }
