@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bfd_if.h - BFD library interface header                                  */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2006                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -36,6 +36,9 @@
 #include <string.h>
 //#include <libgen.h>
 #include <errno.h>
+#ifdef USE_ELFUTILS
+#include <elfutils/libdwfl.h>
+#endif
 #include "avltree.h"
 
 struct sect_cache {
@@ -60,9 +63,13 @@ struct bfd_if {
 	struct disassemble_info	info;
 	struct sect_cache	cache;
 	int			has_debuginfo;
-	bfd_vma			load_vaddr;
+	unsigned long		load_vaddr;
 	int			*p_code_sects;
 	int			n_code_sects;
+
+#ifdef USE_ELFUTILS
+	Dwfl			*dwfl;
+#endif
 };
 
 enum branch_type {
@@ -91,62 +98,64 @@ enum branch_type {
 #define CHK_BTYPE_E_INTERNAL	(1 << 4)
 
 struct unknown {
-	bfd_vma		addr;
+	unsigned long	addr;
 	unsigned long	cnt;
 	struct unknown	*next;
 };
 
 struct path {
-	bfd_vma		addr;		/* state start address */
+	unsigned long	addr;		/* state start address */
 	long		cnt;		/* -1 means next and jmp not linked */
 	int		type;
-	bfd_vma		base;		/* branch/jmp/call instruction address*/
-	bfd_vma		next;		/* fallthrough address */
+	unsigned long	base;		/* branch/jmp/call instruction address*/
+	unsigned long	next;		/* fallthrough address */
 	long		next_cnt;	/* -1 means switch case jump */
-	bfd_vma		jmp;		/* branch/jmp/call destination address
+	unsigned long	jmp;		/* branch/jmp/call destination address
 					 * if jmp == UNKNOWN_BADDR then jmp_cnt
 					 * is pointer to 'struct unknown'
 					 */
 	long		jmp_cnt;
 };
 
-#define get_bfd_vma_fmt(addr)					\
-	sizeof(addr) == sizeof(unsigned long) ? "0x%08lx" : "0x%08llx"
-
-#define	printf_bfd_vma(addr)					\
-	printf(get_bfd_vma_fmt(addr), (addr))
-
-#define	fprintf_err_bfd_vma(addr)					\
-	fprintf(stderr, get_bfd_vma_fmt(addr), (addr))
-
-typedef int (*t_func_for_fsym)(bfd_vma, const char*, void*);
+typedef int (*t_func_for_fsym)(unsigned long, const char*, void*);
 
 
 /*------ proto-type ------*/
 /* initialize */
 int init_bfd_if(struct bfd_if*, const char*, const char*);
-bfd_vma get_offset_addr(struct bfd_if*, bfd_vma);
-void get_kernel_min_max_addr(struct bfd_if*, bfd_vma*, bfd_vma*);
-void remove_useless_fsyms(struct bfd_if*, bfd_vma, bfd_vma, bfd_vma);
+unsigned long get_offset_addr(struct bfd_if*, unsigned long);
+void get_kernel_min_max_addr(struct bfd_if*, unsigned long*, unsigned long*);
+void remove_useless_fsyms(struct bfd_if*, unsigned long, unsigned long,
+			  unsigned long);
 /* path-tree */
-int chk_path_tree(struct bfd_if*, struct path***, long*, bfd_vma);
-struct path* find_path_from_addr(struct path**, long*, bfd_vma);
+int chk_path_tree(struct bfd_if*, struct path***, long*, unsigned long, int);
+int chk_kernel_has_ud2_src_info(struct bfd_if*);
+struct path* find_path_from_addr(struct path**, long*, unsigned long);
+void free_path(struct path*);
 void printf_path(struct path*);
 void dump_path_tree(struct path**, long);
 /* utility */
-int get_source_info(struct bfd_if*, bfd_vma, const char**, const char**, int*);
-int addr_to_func_name_and_offset(struct bfd_if*, bfd_vma, const char**,size_t*);
+#ifdef USE_ELFUTILS
+int init_dwarf(struct bfd_if *bi, const char *obj_name);
+#endif
+int get_source_info(struct bfd_if*, unsigned long, const char**, const char**,
+		    int*);
+int addr_to_func_name_and_offset(struct bfd_if*, unsigned long, const char**,
+				 size_t*);
 void printf_func_name(const char*, size_t);
-int printf_mnemonic(struct bfd_if*, bfd_vma, bfd_vma*);
-int get_branch_info(struct bfd_if*, bfd_vma, int*, bfd_vma*, bfd_vma*, bfd_vma);
-int get_addr_of_symbol(struct bfd_if*, bfd_vma, bfd_vma, bfd_vma,
-		       const char*, bfd_vma*);
-int get_addr_of_symbol_all(struct bfd_if*, bfd_vma, bfd_vma, bfd_vma,
-			   const char*, bfd_vma**);
-int get_end_of_func(struct bfd_if*, bfd_vma, bfd_vma*);
-int for_each_fsymbols(struct bfd_if *bi, bfd_vma offset, t_func_for_fsym f,
-		      void *data);
-void dump_bfd_symbols(struct bfd_if*, bfd_vma, bfd_vma, bfd_vma);
+int printf_mnemonic(struct bfd_if*, unsigned long, unsigned long*);
+int get_branch_info(struct bfd_if*, unsigned long, int*, unsigned long*,
+		    unsigned long*, unsigned long);
+int get_addr_of_symbol(struct bfd_if*, unsigned long, const char*,
+		       unsigned long*);
+int get_addr_of_symbol_all(struct bfd_if*, unsigned long, const char*,
+			   unsigned long**);
+int get_begin_of_func(struct bfd_if*, unsigned long, unsigned long*);
+int get_end_of_func(struct bfd_if*, unsigned long, unsigned long*);
+int for_each_fsymbols(struct bfd_if *bi, unsigned long offset,
+		      t_func_for_fsym f, void *data);
+void dump_bfd_symbols(struct bfd_if*, unsigned long, unsigned long,
+		      unsigned long);
 #ifdef RELOC_TEST
 void enable_reloc_test(void);
 #endif
