@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_ar_parse.c - /proc/modules and /proc/PID/maps parser                  */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -60,6 +60,30 @@ void dump_r2n(struct r2n_info *r2i)
 	}
 }
 
+static void free_path_tree(struct range_to_name *r2n)
+{
+	long i;
+
+	if (!r2n || !r2n->pt)
+		return;
+	for (i = 0; i < r2n->cnt; i++)
+		free_path(r2n->pt[i]);
+	free(r2n->pt);
+	r2n->pt = NULL;
+	r2n->cnt = 0;
+}
+
+static void free_fix_from_cache(struct range_to_name *r2n)
+{
+	struct fix_from_cache *p, *p_next;
+
+	for (p = r2n->fix_from_cache; p; p = p_next) {
+		p_next = p->next;
+		free(p);
+	}
+	r2n->fix_from_cache = NULL;
+}
+
 void free_r2n(struct r2n_info *r2i)
 {
 	int i;
@@ -71,12 +95,36 @@ void free_r2n(struct r2n_info *r2i)
 		p = r2i->all_r2n[i];
 		if (p) {
 			//free(p->func_names);
+			free_path_tree(p);
+			free_fix_from_cache(p);
 			free(p);
 		}
 	}
 	free(r2i->all_r2n);
 }
 
+/* This function is used for merging the kernel coverage. */
+void free_r2n_but_kernel(struct r2n_info *r2i)
+{
+	int i;
+	struct range_to_name *p;
+
+	if (!r2i->all_r2n)
+		return;
+	for (i = 0; i < r2i->r2n_max; i++) {
+		p = r2i->all_r2n[i];
+		if (p) {
+			if (is_kernel(p))
+				continue;
+			//free(p->func_names);
+			free_path_tree(p);
+			free_fix_from_cache(p);
+			free(p);
+			r2i->all_r2n[i] = NULL;
+		}
+	}
+}
+
 struct range_to_name* addr_to_r2n(struct r2n_info *r2i, unsigned long addr)
 {
 	int i;
@@ -182,7 +230,7 @@ static int read_pt_from_file(struct range_to_name *r2n, char *vmlinux_path,
 	if (read(fd, ver, len) != len)
 		goto READ_ERR_EXIT;
 	ver[len] = '\0';
-	if (strcmp(BT_VER, ver) != 0)
+	if (strcmp(VERSION, ver) != 0)
 		goto ERR_EXIT;
 	read_dt(fd, len);
 	if (read(fd, path, len) != len)
@@ -237,9 +285,9 @@ static int write_pt_to_file(struct range_to_name *r2n, char *vmlinux_path,
 	 */
 	chmod(fpath, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
 
-	len = strlen(BT_VER);
+	len = strlen(VERSION);
 	write_dt(fd, len);
-	if (write(fd, BT_VER, len) != len)
+	if (write(fd, VERSION, len) != len)
 		goto ERR_EXIT;
 	len = strlen(vmlinux_path);
 	write_dt(fd, len);
@@ -259,11 +307,51 @@ ERR_EXIT:
 	return -1;
 }
 
+static int get_kernel_offset(struct range_to_name *r2n, char *kallsyms)
+{
+	char *ksym_name = "_stext";
+	unsigned long addr1, addr2;
+	FILE *f = NULL;
+	char buf[MAX_LINE_LEN + 1], type, sym[MAX_LINE_LEN];
+	int found, num;
+
+	if (get_addr_of_symbol(&r2n->bi, 0, ksym_name, &addr1) < 1){
+		fprintf(stderr, "Symbol '%s' not found in kernel.\n",
+			ksym_name);
+		return -1;
+	}
+	if (kallsyms == NULL) {
+		fprintf(stderr, "kallsyms was not logged.\n");
+		return -1;
+	}
+	if ((f = fopen(kallsyms, "r")) == NULL) {
+		fprintf(stderr, "fopen \"%s\" failed.(%s)\n",
+			kallsyms, strerror(errno));
+		return -1;
+	}
+	buf[MAX_LINE_LEN] = '\0';
+	found = 0;
+	while (fgets(buf, MAX_LINE_LEN, f)) {
+		num = sscanf(buf, "%lx %c %s\n", &addr2, &type, sym);
+		if (num == 3 || strcmp(sym, ksym_name) == 0) {
+			found = 1;
+			break;
+		}
+	}
+	fclose(f);
+	if (!found) {
+		fprintf(stderr, "Symbol '%s' not found in '%s'.\n",
+			ksym_name, kallsyms);
+		return -1;
+	}
+	r2n->offset = addr2 - addr1;
+	return 0;
+}
+
 static int prepare_obj_file(struct range_to_name *r2n, char *kallsyms,
 			    char *uname_r, int with_pt, int verbose)
 {
 	char path[MAX_LINE_LEN];
-	int is_kernel = 0;
 	struct stat st;
 
 	snprintf(path, MAX_LINE_LEN, "%s/%s", r2n->dirname, r2n->basename);
@@ -275,51 +363,59 @@ static int prepare_obj_file(struct range_to_name *r2n, char *kallsyms,
 			path, strerror(errno));
 		return -1;
 	}
-	if (init_bfd_if(&r2n->bi, strdup(path), kallsyms) < 0)
+	if (init_bfd_if(&r2n->bi, strdup(path),
+			r2n->is_kernel ? NULL : kallsyms) < 0)
 		return -1;
 	if (!r2n->begin && !r2n->end) {
 		struct addr_range *range = get_pid_addr_range(ALL_PID);
-		bfd_vma min, max;
+		unsigned long min, max;
 
-		is_kernel = 1;
+#ifdef USE_ELFUTILS
+		if (init_dwarf(&r2n->bi, (const char*)path) < 0)
+			return -1;
+#endif
+		if (get_kernel_offset(r2n, kallsyms) < 0)
+			return -1;
 		get_kernel_min_max_addr(&r2n->bi, &min, &max);
 		if (is_addr_range_match(min, max, range)) {
-			r2n->begin = min;
-			r2n->end = max;
+			r2n->begin = min + r2n->offset;
+			r2n->end = max + r2n->offset;
 		} else
 			return 1;
 	} else {
 		r2n->offset = get_offset_addr(&r2n->bi, r2n->begin);
 		/* for DEBUG
-		printf("OFFSET: ");
-		printf_bfd_vma(r2n->offset);
-		printf("\n");
+		printf("OFFSET: 0x%08lx\n", r2n->offset);
 		*/
 	}
 	remove_useless_fsyms(&r2n->bi, r2n->begin, r2n->end, r2n->offset);
 	if (with_pt) {
-		if (is_kernel) {
+		if (is_kernel(r2n)) {
 			if (read_pt_from_file(r2n, path, &st, uname_r) >= 0)
 				return 0;
+			r2n->skip_ud2_srcinfo =
+				chk_kernel_has_ud2_src_info(&r2n->bi);
+			/*
+			printf("Kernel has src-info with ud2=>%d\n",
+			       r2n->skip_ud2_srcinfo);
+			       */
 		}
-		if (chk_path_tree(&r2n->bi, &r2n->pt, &r2n->cnt,r2n->offset) <0)
+		if (chk_path_tree(&r2n->bi, &r2n->pt, &r2n->cnt, r2n->offset,
+				  r2n->skip_ud2_srcinfo) < 0)
 			return -1;
-		if (is_kernel)
+		if (is_kernel(r2n))
 			return write_pt_to_file(r2n, path, &st, uname_r);
 	}
 	return 0;
 }
 
-static int add_range_to_name(struct r2n_info *r2i,
-			     unsigned long begin, unsigned long end,
-			     char *name, char *kallsyms, char *uname_r,
-			     int is_module, int with_pt, int verbose)
+int add_range_to_name(struct r2n_info *r2i,
+		      unsigned long begin, unsigned long end, char *name,
+		      char *kallsyms, int is_module, int with_pt, int verbose)
 {
-	int rc, i, step = 128, r2n_max;
-	struct range_to_name *r2n, **all_r2n;
+	int rc, i, step = 8;
+	struct range_to_name *r2n;
 
-	all_r2n = r2i->all_r2n;
-	r2n_max = r2i->r2n_max;
 	if (is_same_name_registered(r2i, name))
 		return 0;
 	r2n = xcalloc(1, sizeof(struct range_to_name));
@@ -328,22 +424,24 @@ static int add_range_to_name(struct r2n_info *r2i,
 	snprintf(r2n->name, PATH_MAX, "%s", name);
 	r2n->basename = basename(r2n->name);
 	r2n->dirname = dirname(r2n->name);
+	r2n->is_kernel = strcmp(r2n->basename, KERNEL) == 0;
 
-	for (i = 0; i < r2n_max && all_r2n[i]; i++);
-	if (i >= r2n_max) {
-		all_r2n = xrealloc(all_r2n, (r2n_max + step) * sizeof(r2n));
-		memset(&all_r2n[i], 0, step * sizeof(r2n));
-		i = r2n_max;
-		r2n_max += step;
+	for (i = 0; i < r2i->r2n_max && r2i->all_r2n[i]; i++);
+	if (i >= r2i->r2n_max) {
+		r2i->all_r2n = xrealloc(r2i->all_r2n,
+					(r2i->r2n_max + step) * sizeof(r2n));
+		memset(&r2i->all_r2n[i], 0, step * sizeof(r2n));
+		i = r2i->r2n_max;
+		r2i->r2n_max += step;
 	}
-	all_r2n[i] = r2n;
-	rc = prepare_obj_file(r2n, kallsyms, uname_r, with_pt, verbose);
+	r2i->all_r2n[i] = r2n;
+	rc = prepare_obj_file(r2n, kallsyms, r2i->uname_r, with_pt, verbose);
 	if (rc < 0)
 		//return 0;
 		return -1;
 	else if (rc > 0) {
 		free(r2n);
-		all_r2n[i] = NULL;
+		r2i->all_r2n[i] = NULL;
 	}
 	return 0;
 }
@@ -823,10 +921,29 @@ EXIT:
 	return rc;
 }
 
+static char *elf_path_prefix = "";
+
+void set_elf_path_prefix(char *prefix)
+{
+	elf_path_prefix = strdup(prefix);
+}
+
+char* get_elf_path_prefix(void)
+{
+	return elf_path_prefix;
+}
+
 static int get_module_full_path(char *buf, int buf_len, const char *name,
 				char *uname_r)
 {
-	snprintf(buf, buf_len, "/lib/modules/%s", uname_r);
+	int len;
+
+	len = snprintf(buf, buf_len, "%s/lib/modules/%s",
+		       elf_path_prefix, uname_r);
+	if (len >= buf_len) {
+		fprintf(stderr, "!!! too short buf size(%d)\n", buf_len);
+		return -1;
+	}
 	return find_under_path(buf, buf_len, name, 1);
 }
 
@@ -834,42 +951,128 @@ static int __get_vmlinux_full_path(char *buf, int buf_len)
 {
 	int rc;
 
-	rc = find_under_path(buf, buf_len, "vmlinux", 0);
+	rc = find_under_path(buf, buf_len, KERNEL, 0);
 	if (rc >= 0)
 		return rc;
-	return find_under_path(buf, buf_len, "vmlinux", 1);
+	return find_under_path(buf, buf_len, KERNEL, 1);
 }
 
 static int get_vmlinux_full_path(char *buf, int buf_len, char *uname_r)
 {
-	snprintf(buf, buf_len, "/usr/lib/debug/lib/modules/%s", uname_r);
+	snprintf(buf, buf_len, "%s/usr/lib/debug/lib/modules/%s",
+		 elf_path_prefix, uname_r);
 	if (__get_vmlinux_full_path(buf, buf_len) >= 0)
 		return 0;
-	snprintf(buf, buf_len, "/lib/modules/%s/build", uname_r);
+	snprintf(buf, buf_len, "%s/lib/modules/%s/build",
+		 elf_path_prefix, uname_r);
 	if (__get_vmlinux_full_path(buf, buf_len) >= 0)
 		return 0;
 	return -1;
 }
 
-int parse_modules(struct r2n_info *r2i, char *dir, char *uname_r,
-		  int with_pt, int verbose)
+static int change_module_name(char *name)
 {
-	FILE *fd;
+	int changed;
+	char *p;
+
+	changed = 0;
+	for (p = name; p && *p; p++) {
+		switch (*p) {
+		case '-':
+		case '_':
+			changed = 1;
+			if (*p == '-')
+				*p = '_';
+			else
+				*p = '-';
+		}
+	}
+	return changed;
+}
+
+static unsigned long irq_addrs[256 + 1];
+static int irq_addrs_cnt;
+
+static int comp_address(const void *p_a1, const void *p_a2)
+{
+	unsigned long a1 = *(unsigned long*)p_a1;
+	unsigned long a2 = *(unsigned long*)p_a2;
+
+	if (a1 < a2)
+		return -1;
+	if (a1 > a2)
+		return 1;
+	return 0;
+}
+
+/* return address count */
+static int get_uniq_and_sorted_irq_addrs(char *dir,
+					 unsigned long addrs[256 + 1])
+{
+	int i, j, already_checked, rc;
+	char buf[PATH_MAX];
+	FILE *f;
+	unsigned long a;
+
+	if (!dir)
+		return 0;
+	rc = -1;
+	snprintf(buf, PATH_MAX, "%s/irq_addrs", dir);
+	if ((f = fopen(buf, "r")) == NULL) {
+		fprintf(stderr, "fopen \"%s\" failed.(%s)\n",
+			buf, strerror(errno));
+		goto EXIT;
+	}
+	i = 0;
+	while (fgets(buf, PATH_MAX, f)) {
+		if (sscanf(buf, "%lx\n", &a) != 1) {
+			fprintf(stderr, "!!! wrong format irq_addrs file\n");
+			goto EXIT;
+		}
+		already_checked = 0;
+		for (j = 0; j < i; j++)
+			if (addrs[j] == a) {
+				already_checked = 1;
+				break;
+			}
+		if (already_checked)
+			continue;
+		addrs[i++] = a;
+		if (i >= 256) {
+			fprintf(stderr, "!!! wrong format irq_addrs file\n");
+			goto EXIT;
+		}
+	}
+	rc = i;
+	qsort(addrs, i, sizeof(unsigned long), comp_address);
+EXIT:
+	if (f)
+		fclose(f);
+	return rc;
+}
+
+static inline int is_irq_addrs(unsigned long a, unsigned long addrs[256 + 1],
+			       int n)
+{
+	return bsearch(&a, addrs, n, sizeof(unsigned long), comp_address)
+			!= NULL;
+}
+#define IS_IRQ(a)	is_irq_addrs((a), irq_addrs, irq_addrs_cnt)
+
+int parse_modules(struct r2n_info *r2i, char *dir, int with_pt, int verbose)
+{
+	FILE *fd = NULL;
 	int rc, num, len;
 	char path[128], kallsyms[128];
 	char buf[MAX_LINE_LEN + 1], *p, fullname[MAX_LINE_LEN + 1];
-	struct utsname utsn;
 	struct addr_range *range;
 	unsigned long from, to;
 
 	rc = -1;
-	if (!uname_r) {
-		if (uname(&utsn) < 0)
-			goto EXIT;
-		uname_r = utsn.release;
-	}
 	if (!dir)
 		goto CHK_KERNEL;
+	if ((irq_addrs_cnt = get_uniq_and_sorted_irq_addrs(dir, irq_addrs)) < 0)
+		goto EXIT;
 	snprintf(kallsyms, sizeof(kallsyms), "%s/kallsyms", dir);
 	snprintf(path, sizeof(path), "%s/modules", dir);
 	if ((fd = fopen(path, "r")) == NULL) {
@@ -892,24 +1095,27 @@ int parse_modules(struct r2n_info *r2i, char *dir, char *uname_r,
 			continue;
 		snprintf(buf, MAX_LINE_LEN, "%s.ko", fullname);
 		fullname[0] = '\0';
-		if (get_module_full_path(fullname, MAX_LINE_LEN, buf, uname_r)
-		    < 0) {
+		if (get_module_full_path(fullname, MAX_LINE_LEN, buf,
+					 r2i->uname_r) < 0 &&
+		    (!change_module_name(buf) ||
+		     get_module_full_path(fullname, MAX_LINE_LEN, buf,
+					  r2i->uname_r) < 0)) {
 			if (verbose)
 				printf("WARN: %s not found.\n", buf);
 			continue;
 		}
-		if (add_range_to_name(r2i, from, to, fullname, kallsyms,
-				      uname_r, 1, with_pt, verbose) < 0)
+		if (add_range_to_name(r2i, from, to, fullname, kallsyms, 1,
+				      with_pt, verbose) < 0)
 			goto EXIT;
 	}
 CHK_KERNEL:
 	fullname[0] = '\0';
-	if (get_vmlinux_full_path(fullname, MAX_LINE_LEN, uname_r) < 0) {
+	if (get_vmlinux_full_path(fullname, MAX_LINE_LEN, r2i->uname_r) < 0) {
 		if (verbose)
 			printf("WARN: vmlinux not found.\n");
 	} else {
-		if (add_range_to_name(r2i, 0, 0, fullname, NULL, uname_r, 0,
-				      with_pt, verbose) < 0)
+		if (add_range_to_name(r2i, 0, 0, fullname, kallsyms, 0, with_pt,
+				      verbose) < 0)
 			goto EXIT;
 	}
 	rc = 0;
@@ -927,15 +1133,15 @@ int get_symbol_addr(struct r2n_info *r2i,
 {
 	int i, found, tmp;
 	struct range_to_name *r2n;
-	bfd_vma __addr;
+	unsigned long __addr;
 
 	if (!r2i->all_r2n)
 		return 0;
 	for (found = 0, i = 0; i < r2i->r2n_max; i++) {
 		r2n = r2i->all_r2n[i];
 		if (r2n) {
-			tmp = get_addr_of_symbol(&r2n->bi, r2n->begin, r2n->end,
-						 r2n->offset, funcsym, &__addr);
+			tmp = get_addr_of_symbol(&r2n->bi, r2n->offset, funcsym,
+						 &__addr);
 			if (!found && tmp)
 				*addr = __addr;
 			found += tmp;
@@ -949,8 +1155,8 @@ int get_symbol_addr_all(struct r2n_info *r2i, const char *funcsym,
 {
 	int i, j, rc, cnt;
 	struct range_to_name *r2n;
-	unsigned long *addrs;
-	bfd_vma *c_addrs;
+	unsigned long *addrs = NULL;
+	unsigned long *c_addrs;
 
 	if (!r2i->all_r2n)
 		return 0;
@@ -958,8 +1164,7 @@ int get_symbol_addr_all(struct r2n_info *r2i, const char *funcsym,
 	for (i = 0; i < r2i->r2n_max; i++) {
 		r2n = r2i->all_r2n[i];
 		if (r2n) {
-			rc = get_addr_of_symbol_all(&r2n->bi, r2n->begin,
-						    r2n->end, r2n->offset,
+			rc = get_addr_of_symbol_all(&r2n->bi, r2n->offset,
 						    funcsym, &c_addrs);
 			if (rc < 0)
 				return -1;
@@ -981,25 +1186,61 @@ int get_symbol_addr_all(struct r2n_info *r2i, const char *funcsym,
 	return cnt;
 }
 
+const char* get_fname(struct range_to_name *r2n, unsigned long addr)
+{
+	const char *fname;
+	size_t offset;
+
+	if (addr_to_func_name_and_offset(&r2n->bi, addr - r2n->offset,
+					 &fname, &offset) < 0)
+		return NULL;
+	return fname;
+}
+
+#if 0
+static int get_func_begin(struct range_to_name *r2n, unsigned long addr,
+			  unsigned long *p_begin)
+{
+	unsigned long begin;
+
+	if (r2n) {
+		if (get_begin_of_func(&r2n->bi, addr - r2n->offset, &begin)) {
+			*p_begin = begin + r2n->offset;
+			return 1;
+		}
+	}
+	return 0;
+}
+#endif
+
+static int get_func_end(struct range_to_name *r2n, unsigned long begin,
+			unsigned long *p_end)
+{
+	unsigned long end;
+
+	if (r2n) {
+		end = r2n->end - r2n->offset;
+		if (get_end_of_func(&r2n->bi, begin - r2n->offset, &end)) {
+			*p_end = end + r2n->offset;
+			return 1;
+		}
+	}
+	return 0;
+}
+
 int get_func_info(struct r2n_info *r2i, unsigned long begin,
 		  unsigned long *p_end, struct range_to_name **p_r2n)
 {
 	int i;
 	struct range_to_name *r2n;
-	bfd_vma end;
 
 	if (!r2i->all_r2n)
 		return 0;
 	for (i = 0; i < r2i->r2n_max; i++) {
 		r2n = r2i->all_r2n[i];
-		if (r2n) {
-			end = r2n->end - r2n->offset;
-			if (get_end_of_func(&r2n->bi, begin - r2n->offset,
-					    &end)) {
-				*p_end = end + r2n->offset;
-				*p_r2n = r2n;
-				return 1;
-			}
+		if (get_func_end(r2n, begin, p_end)) {
+			*p_r2n = r2n;
+			return 1;
 		}
 	}
 	return 0;
@@ -1020,7 +1261,7 @@ int parse_maps(struct r2n_info *r2i, char *dir, char *fname,
 	FILE *fd;
 	int rc, num, len;
 	char path[128];
-	char mode[6], buf[MAX_LINE_LEN + 1], *p;
+	char mode[6], buf[MAX_LINE_LEN + 1], *p, fullname[MAX_LINE_LEN + 1];
 	struct addr_range *range;
 	unsigned long from, to;
 	long size;
@@ -1031,7 +1272,7 @@ int parse_maps(struct r2n_info *r2i, char *dir, char *fname,
 		fprintf(stderr, "can't open %s.(%s)\n", path, strerror(errno));
 		goto EXIT;
 	}
-	buf[MAX_LINE_LEN] = '\0';
+	buf[MAX_LINE_LEN] = fullname[MAX_LINE_LEN] = '\0';
 	range = get_pid_addr_range(ALL_PID);
 	while ((p = fgets(buf, MAX_LINE_LEN, fd)) != NULL) {
 		len = strlen(p) - 1;
@@ -1050,8 +1291,9 @@ int parse_maps(struct r2n_info *r2i, char *dir, char *fname,
 		to -= 1;
 		if (!is_addr_range_match(from, to, range))
 			continue;
-		if (add_range_to_name(r2i, from, to, p, NULL, NULL, 0,
-				      with_pt, verbose) < 0)
+		snprintf(fullname, MAX_LINE_LEN, "%s%s", elf_path_prefix, p);
+		if (add_range_to_name(r2i, from, to, fullname, NULL, 0, with_pt,
+				      verbose) < 0)
 			goto EXIT;
 	}
 	rc = 0;
@@ -1070,3 +1312,326 @@ int chk_from_is_next(char *dir)
 	return (stat(path, &st) == 0);
 }
 
+int set_uname(struct r2n_info *r2i, char *log_path)
+{
+	struct utsname utsn;
+	int max_uname_len = 80;
+	char *dir, *tmp, path[PATH_MAX], uname_str[max_uname_len], *p;
+	FILE *f;
+
+	if (r2i->uname_r)
+		return 0;
+	if (log_path) {
+		tmp = strdup(log_path);
+		dir = dirname(tmp);
+		snprintf(path, PATH_MAX, "%s/uname", dir);
+		free(tmp);
+		if ((f = fopen(path, "r"))) {
+			p = fgets(uname_str, max_uname_len, f);
+			fclose(f);
+			if (p) {
+				p[strlen(p) - 1] = '\0';
+				r2i->uname_r = strdup(p);
+				return 0;
+			}
+		}
+	}
+	if (uname(&utsn) < 0) {
+		fprintf(stderr, "uname failed.(%s)\n", strerror(errno));
+		return -1;
+	}
+	r2i->uname_r = strdup(utsn.release);
+	return 0;
+}
+
+/*----------------------------------------------------------------------------
+ *  check enter to and leave from context (function-call and interrupt) 
+ *----------------------------------------------------------------------------
+ */
+#define MAX_NEST	2048
+
+struct ret_chk {
+	int			type;
+	unsigned long		ret_addr;
+};
+
+struct st_ret_chk {
+	int		nest;
+	struct ret_chk	ret_chk[MAX_NEST];
+};
+
+static struct st_ret_chk st_ret_chk;
+
+#ifdef DEBUG
+static void dump_ret_chk(struct st_ret_chk *retc, int n)
+{
+	int i;
+	struct ret_chk *p;
+
+	for (i = n; i < retc->nest; i++) {
+		p = &retc->ret_chk[i];
+		printf("%04d:0x%08lx(%d)\n", n, p->ret_addr, p->type);
+	}
+}
+#endif
+
+static void set_expect_ret_addr(struct st_ret_chk *retc, int type,
+				unsigned long ret_addr)
+{
+	struct ret_chk *p;
+
+	if (retc->nest >= MAX_NEST) {
+		printf("WARN: MAX_NEST is too small.\n");
+		return;
+	}
+	p = &retc->ret_chk[retc->nest];
+	p->type = type;
+	p->ret_addr = ret_addr;
+	retc->nest++;
+}
+
+static void get_nest_from_ret_addr(struct st_ret_chk *retc, unsigned long addr,
+				   int ret_maybe_fixup)
+{
+	int i;
+	struct ret_chk *p;
+
+	for (i = retc->nest - 1; i >= 0; i--) {
+		p = &retc->ret_chk[i];
+		if (ret_maybe_fixup) {
+			if (p->type == BTYPE_INT) {
+				retc->nest = i;
+				return;
+			}
+		} else {
+			if (p->ret_addr == addr) {
+				retc->nest = i;
+				return;
+			}
+		}
+	}
+	/* If there is trace-off and trace-on, 'ret_chk' data is almost meaning
+	 * less because real stack was already changed.
+	 * But normally, trace-off and trace-on's both nest are the same.
+	 * So, if we couldn't find the return address in 'ret_chk', then
+	 * simply minus one from nest value.
+	 */
+	if (retc->nest)
+		retc->nest--;
+	return;
+}
+
+static int proc_enter(struct st_ret_chk *retc, int type, int nest_chk,
+		      struct bt_record *p, unsigned long ret)
+{
+	switch (type) {
+	case BTYPE_CALL:
+	case BTYPE_INT:
+	case BTYPE_BREAK:
+		/*
+		printf("ENTER(%d): %d, from:%08lx, to:%08lx, ret:%08lx\n",
+		       retc->nest, type, p->from, p->to, ret);
+		       */
+		set_expect_ret_addr(retc, type, ret);
+		return 2;
+	}
+	return 1;
+}
+
+static int proc_leave(struct st_ret_chk *retc, int type, int nest_chk,
+		      struct bt_record *p, int ret_maybe_fixup)
+{
+	switch (type) {
+	case BTYPE_CALL:
+	case BTYPE_INT:
+	case BTYPE_BREAK:
+		/*
+		printf("LEAVE(%d): %d, from:%08lx, to:%08lx\n",
+		       retc->nest, type, p->from, p->to);
+		       */
+		get_nest_from_ret_addr(retc, p->to, ret_maybe_fixup);
+		return 3;
+	}
+	return 1;
+}
+
+void chk_nest_initialize(void)
+{
+	st_ret_chk.nest = 0;
+}
+
+static inline int dest_is_kernel_func(struct range_to_name *r2n,
+				      struct bt_record *p)
+{
+	const char *fname;
+	size_t offset;
+
+	return (is_kernel(r2n) &&
+		(addr_to_func_name_and_offset(&r2n->bi, p->to - r2n->offset,
+					      &fname, &offset) >= 0) &&
+		fname && offset == 0);
+}
+
+#define CALL(type, nest_chk, p, p_from, rc) \
+	do { \
+		*(type) = BTYPE_CALL; \
+		(rc) = proc_enter(&st_ret_chk, *(type), (nest_chk), (p), \
+				  (p_from)->next); \
+	} while (0)
+
+#define RET(type, nest_chk, p, rc) \
+	do { \
+		*(type) = BTYPE_CALL; \
+		(rc) = proc_leave(&st_ret_chk, *(type), (nest_chk), (p), 0); \
+	} while (0)
+
+#define INT(type, nest_chk, p, rc) \
+	do { \
+		*(type) = BTYPE_INT; \
+		(rc) = proc_enter(&st_ret_chk, *(type), (nest_chk), (p), \
+				  (p)->from); \
+	} while (0)
+
+#define IRET(type, nest_chk, p, ret_maybe_fixup, rc) \
+	do { \
+		*(type) = BTYPE_INT; \
+		(rc) = proc_leave(&st_ret_chk, *(type), (nest_chk), (p), \
+				  (ret_maybe_fixup)); \
+	} while (0)
+
+#define JMP(type, nest_chk, p, rc) \
+	do { \
+		*(type) = BTYPE_JMP; \
+		(rc) = proc_enter(&st_ret_chk, *(type), (nest_chk), (p), \
+				  UNKNOWN_BADDR); \
+	} while (0)
+
+/* return value =  0 : not process ('from' and 'to' are both out of range)
+ *                 1 : not process (normal branch or jump etc.)
+ *                 2 : enter
+ *                 3 : leave
+ *                -1 : error
+ */
+int chk_enter_leave(struct r2n_info *r2i, struct bt_record *p,
+		    int nest_chk, int *inst_type, int *type, int *nest,
+		    struct range_to_name **p_r2n_from,
+		    struct range_to_name **p_r2n_to,
+		    struct path **pp_from, long *p_idx_from)
+{
+	int rc, ret_maybe_fixup = 0;
+	long idx_from = -1;
+	struct range_to_name *r2n = NULL, *r2n_to = NULL;
+	struct path *p_from = NULL;
+
+	*nest = st_ret_chk.nest;
+	r2n_to = addr_to_r2n(r2i, p->to);
+	rc = chk_fix_from_cache(r2i, &p->from, p->to, &r2n, &p_from, &idx_from);
+	if (rc < 0) {
+		*inst_type = BTYPE_OTHER;
+		goto EXIT;
+	}
+	/*
+	if (!is_addr_range_match(p->from, p->to, data->range))
+		return 0;
+		*/
+	if (rc == 0) {
+		/* Branch from unknown-binary */
+		if (!r2n_to)
+			goto EXIT;
+		if (is_kernel(r2n_to)) {
+			*inst_type = BTYPE_OTHER;
+			if (IS_IRQ(p->to))
+				INT(type, nest_chk, p, rc);
+			else
+				IRET(type, nest_chk, p, ret_maybe_fixup, rc);
+		} else {
+			/* Branch from unknown-binary to user-land means
+			 * iret or ret (from 'vsdo').
+			 * We assume it iret because we can't identify these.
+			 */
+			*inst_type = BTYPE_OTHER;
+			IRET(type, nest_chk, p, ret_maybe_fixup, rc);
+		}
+		goto EXIT;
+	}
+	*inst_type = p_from->type;
+	if (p_from->base != p->from)
+		*inst_type = BTYPE_OTHER;
+
+	if (!r2n_to) {
+		/* Branch to unknown-binary */
+		if (is_kernel(r2n))
+			IRET(type, nest_chk, p, ret_maybe_fixup, rc);
+		else {
+			if (IS_IRQ(p->to))
+				INT(type, nest_chk, p, rc);
+			else
+				CALL(type, nest_chk, p, p_from, rc);
+		}
+		goto EXIT;
+	}
+
+	switch (*inst_type) {
+	case BTYPE_BRANCH:
+	case BTYPE_JMP:
+	case BTYPE_CALL:
+		if ((p_from->jmp != UNKNOWN_BADDR && p_from->jmp == p->to) ||
+		    (p_from->jmp == UNKNOWN_BADDR && !IS_IRQ(p->to))) {
+			switch (*inst_type) {
+			case BTYPE_JMP:
+				if (dest_is_kernel_func(r2n_to, p))
+					CALL(type, nest_chk, p, p_from, rc);
+				else
+					JMP(type, nest_chk, p, rc);
+				break;
+			case BTYPE_CALL:
+				CALL(type, nest_chk, p, p_from, rc);
+				break;
+			case BTYPE_BRANCH:
+				rc = 1;		/* do nothing */
+				break;
+			}
+		} else 
+			INT(type, nest_chk, p, rc);
+		goto EXIT;
+	case BTYPE_RET:
+		if (IS_IRQ(p->to)) {
+			INT(type, nest_chk, p, rc);
+		} else if (!(is_kernel(r2n) || is_kernel(r2n_to)) &&
+			   p_from->next - p->from > 1) {
+			/* 'ret n' from user-land to user-land means jump to
+			 * the plt library function.
+			 */
+			JMP(type, nest_chk, p, rc);
+		} else
+			RET(type, nest_chk, p, rc);
+		goto EXIT;
+	case BTYPE_INT:
+	case BTYPE_BREAK:
+	case BTYPE_OTHER:
+		INT(type, nest_chk, p, rc);
+		goto EXIT;
+	case BTYPE_IRET:
+		/* If the 'fixup_exception' function executed, then iret address
+		 * was changed.
+		 * So, in this case, we will not check the return address.
+		 */
+		ret_maybe_fixup = r2n == r2n_to; /*iret from kernel to kernel?*/
+		IRET(type, nest_chk, p, ret_maybe_fixup, rc);
+		goto EXIT;
+	}
+	rc = 0;
+EXIT:
+	if (rc == 3) /* leave? */
+		*nest = st_ret_chk.nest;
+	if (p_r2n_from)
+		*p_r2n_from = r2n;
+	if (p_r2n_to)
+		*p_r2n_to = r2n_to;
+	if (pp_from)
+		*pp_from = p_from;
+	if (p_idx_from)
+		*p_idx_from = idx_from;
+	return rc;
+}
+
