@@ -5,8 +5,8 @@
 
 /*****************************************************************************/
 /*  bt_ar_parse.h - /proc/modules and /proc/PID/maps parse header            */
-/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2007                         */
-/*             Authors: Yumiko Sugita (sugita@sdl.hitachi.co.jp),            */
+/*  Copyright: Copyright (c) Hitachi, Ltd. 2005-2008                         */
+/*             Authors: Yumiko Sugita (yumiko.sugita.yf@hitachi.com),        */
 /*                      Satoshi Fujiwara (sa-fuji@sdl.hitachi.co.jp)         */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
@@ -33,6 +33,7 @@
 #include <limits.h>
 
 #define	ALL_PID			-1
+#define KERNEL			"vmlinux"
 
 struct addr_range {
 	unsigned long		begin;
@@ -64,6 +65,8 @@ struct range_to_name {
 	char			name[PATH_MAX];
 	char			*dirname;
 	char			*basename;
+	int			is_kernel;
+	int			skip_ud2_srcinfo;
 
 	struct bfd_if		bi;
 	struct path		**pt;
@@ -71,11 +74,13 @@ struct range_to_name {
 
 	struct fix_from_cache	*fix_from_cache;
 };
+#define is_kernel(r2n)	(r2n)->is_kernel
 
 struct r2n_info {
 	struct range_to_name	**all_r2n;
 	int			r2n_max;
 	int			from_is_next;
+	char			*uname_r;
 };
 
 #define eprintf(...)	fprintf(stderr, __VA_ARGS__)
@@ -87,6 +92,9 @@ struct r2n_info {
 #define ddprintf(...)
 #endif
 
+int add_range_to_name(struct r2n_info *r2i,
+		      unsigned long begin, unsigned long end, char *name,
+		      char *kallsyms, int is_module, int with_pt, int verbose);
 int alloc_pid_range(pid_t);
 void free_ranges(void);
 char* range2ulongs(char*, unsigned long*, unsigned long*);
@@ -98,6 +106,7 @@ struct addr_range *get_pid_addr_range(pid_t);
 int is_addr_range_match(unsigned long, unsigned long, struct addr_range*);
 int get_symbol_addr(struct r2n_info*, const char*, unsigned long*);
 int get_symbol_addr_all(struct r2n_info*, const char*, unsigned long**);
+const char* get_fname(struct range_to_name*, unsigned long);
 int get_func_info(struct r2n_info*, unsigned long, unsigned long*,
 		  struct range_to_name**);
 
@@ -107,8 +116,9 @@ int is_exists_maps(char*, char*);
 #define PM_WITH_PT	1
 #define PM_SILENT 	0
 #define PM_VERBOSE	1
-int parse_modules(struct r2n_info*, char *dir, char *uname_r,
-		  int with_pt, int verbose);
+void set_elf_path_prefix(char *prefix);
+char* get_elf_path_prefix(void);
+int parse_modules(struct r2n_info*, char *dir, int with_pt, int verbose);
 int parse_maps(struct r2n_info*, char *dir, char *fname,
 	       int with_pt, int verbose);
 
@@ -117,8 +127,18 @@ int chk_from_is_next(char*);
 
 void dump_r2n(struct r2n_info*);
 void free_r2n(struct r2n_info*);
+void free_r2n_but_kernel(struct r2n_info*);
 struct range_to_name* addr_to_r2n(struct r2n_info*, unsigned long);
 int chk_fix_from_cache(struct r2n_info*, unsigned long*, unsigned long,
 		       struct range_to_name**, struct path**, long*);
+int set_uname(struct r2n_info*, char*);
+
+/* checking enter to and leave from functions or interrupt */
+void chk_nest_initialize(void);
+int chk_enter_leave(struct r2n_info *r2i, struct bt_record *p,
+		    int nest_chk, int *inst_type, int *type, int *nest,
+		    struct range_to_name **p_r2n_from,
+		    struct range_to_name **p_r2n_to,
+		    struct path **pp_from, long *p_idx_from);
 
 #endif /* __BT_AR_PARSE_H__ */
